from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, model_validator, ConfigDict
from pydantic_extra_types.currency_code import Currency

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumTransactionType, EnumTransactionState


class ExchangeMoney:
    """
    Exchange money using one of these methods:

    Sell currency:
        You know the amount of currency to sell.
        For example, you want to exchange 135.5 USD to some EUR.
        Specify the amount in the from object.

    Buy currency:
        You know the amount of currency to buy.
        For example, you want to exchange some USD to 200 EUR.
        Specify the amount in the to object.
    """

    ROUTE = "/1.0/exchange"

    class Body(BaseModel):
        """
        The request body for the endpoint.
        """

        model_config = ConfigDict(
            populate_by_name=True,
            from_attributes=True,
        )

        class ModelFrom(BaseModel):
            """The details of the currency to exchange from."""

            account_id: Annotated[
                UUID, Field(description="The ID of the account to sell currency from.")
            ]
            currency: Annotated[
                Currency,
                Field(description="The currency to sell in ISO 4217 format."),
            ]
            amount: Annotated[
                float | None,
                Field(
                    description="The amount of currency. Specify ONLY if you want to sell currency."
                ),
            ] = None

        class ModelTo(BaseModel):
            """The details of the currency to exchange to."""

            account_id: Annotated[
                UUID,
                Field(
                    description="The ID of the account to receive exchanged currency into."
                ),
            ]
            currency: Annotated[
                Currency,
                Field(description="The currency to buy in ISO 4217 format."),
            ]
            amount: Annotated[
                float | None,
                Field(
                    description="The amount of currency. Specify ONLY if you want to buy currency."
                ),
            ] = None

        from_: Annotated[
            ModelFrom,
            Field(
                alias="from",
                description="The details of the currency to exchange from.",
            ),
        ]
        to: Annotated[
            ModelTo,
            Field(description="The details of the currency to exchange to."),
        ]
        reference: Annotated[
            str | None,
            Field(
                description="""
                The reference for the exchange transaction, provided by you. 
                It helps you to identify the transaction if you want to look it up later.
                """
            ),
        ] = None
        request_id: Annotated[
            str,
            Field(
                description="""
                The ID of the request, provided by you. 
                It helps you identify the transaction in your system.

                To ensure that an exchange transaction is not processed multiple 
                times if there are network or system errors, the same request_id 
                should be used for requests related to the same transaction.
                """,
                max_length=40,
            ),
        ]

        @model_validator(mode="after")
        def check_inputs(self) -> "ExchangeMoney.Body":
            """Check if the amount is specified in either the from or to object."""
            if self.from_.amount is None and self.to.amount is None:
                raise ValueError(
                    "Either the amount in the from or to object must be specified."
                )
            if self.from_.amount is not None and self.to.amount is not None:
                raise ValueError(
                    "Only the amount in either the from or to object must be specified."
                )
            return self

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        id: Annotated[
            UUID | None, Field(description="The ID of the created transaction.")
        ] = None
        type: Annotated[
            EnumTransactionType,
            Field(
                description="The type of the transaction. For money exchange, it is 'exchange'."
            ),
        ] = EnumTransactionType.EXCHANGE
        reason_code: Annotated[
            str | None,
            Field(
                description="The reason code when the state parameter of the transaction is declined or failed."
            ),
        ] = None
        created_at: Annotated[
            DateTime,
            Field(
                description="The date and time the transaction was created in ISO 8601 format."
            ),
        ]
        completed_at: Annotated[
            DateTime | None,
            Field(
                description="The date and time the transaction was completed in ISO 8601 format."
            ),
        ] = None
        state: Annotated[
            EnumTransactionState | None,
            Field(
                description="""
                Indicates the transaction state. Possible values:

                    created:
                        The transaction has been created and is either processed asynchronously
                        or scheduled for a later time.
                    pending:
                        The transaction is pending until it's being processed.
                        If the transfer is made between Revolut accounts,
                        this state is skipped and the transaction is executed instantly.
                    completed:
                        The transaction was successful.
                    declined:
                        The transaction was unsuccessful. This can happen for a variety of reasons,
                        for example, insufficient account balance, wrong receiver information, etc.
                    failed:
                        The transaction was unsuccessful. This can happen for a variety of reasons,
                        for example, invalid API calls, blocked payments, etc.
                    reverted:
                        The transaction was reverted. This can happen for a variety of reasons,
                        for example, the receiver being inaccessible.
                """
            ),
        ] = None



from .common import PyRevolutBaseException


class PyRevolutNetworkError(PyRevolutBaseException):
    """Network Error -- The request failed due to a network error."""

    def __init__(
        self,
        msg="The request failed due to a network error. Please check your network connection and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from uuid import UUID

from pyrevolut.api.common import BaseEndpointAsync

from pyrevolut.api.foreign_exchange.get import GetExchangeRate
from pyrevolut.api.foreign_exchange.post import ExchangeMoney


class EndpointForeignExchangeAsync(BaseEndpointAsync):
    """The async Foreign Exchange API

    Retrieve information on exchange rates between currencies, buy and sell currencies.
    """

    async def get_exchange_rate(
        self,
        from_currency: str,
        to_currency: str,
        amount: float | None = None,
        **kwargs,
    ) -> dict | GetExchangeRate.Response:
        """
        Get the sell exchange rate between two currencies.

        Parameters
        ----------
        from_currency : str
            The currency that you exchange from in ISO 4217 format.
        to_currency : str
            The currency that you exchange to in ISO 4217 format.
        amount : float | None
            The amount of the currency to exchange from.
            The default value is 1.00 if not provided.

        Returns
        -------
        dict | GetExchangeRate.Response
            A dict with the information about the exchange rate.
        """
        endpoint = GetExchangeRate
        path = endpoint.ROUTE
        params = endpoint.Params(
            from_=from_currency,
            to=to_currency,
            amount=amount,
        )

        return await self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    async def exchange_money(
        self,
        request_id: str,
        from_account_id: UUID,
        from_currency: str,
        to_account_id: UUID,
        to_currency: str,
        from_amount: float | None = None,
        to_amount: float | None = None,
        reference: str | None = None,
        **kwargs,
    ) -> dict | ExchangeMoney.Response:
        """
        Exchange money using one of these methods:

        Sell currency:
            You know the amount of currency to sell.
            For example, you want to exchange 135.5 USD to some EUR.
            Specify the amount in the from object.

        Buy currency:
            You know the amount of currency to buy.
            For example, you want to exchange some USD to 200 EUR.
            Specify the amount in the to object.

        Parameters
        ----------
        request_id : str
            The ID of the request, provided by you.
            It helps you identify the transaction in your system.

            To ensure that an exchange transaction is not processed multiple
            times if there are network or system errors, the same request_id
            should be used for requests related to the same transaction.
        from_account_id : UUID
            The ID of the account to sell currency from.
        from_currency : str
            The currency to sell in ISO 4217 format.
        to_account_id : UUID
            The ID of the account to receive exchanged currency into.
        to_currency : str
            The currency to buy in ISO 4217 format.
        from_amount : float | None
            The amount of currency. Specify ONLY if you want to sell currency.
        to_amount : float | None
            The amount of currency. Specify ONLY if you want to buy currency.
        reference : str | None
            The reference for the exchange transaction, provided by you.
            It helps you to identify the transaction if you want to look it up later.

        Returns
        -------
        dict | ExchangeMoney.Response
            A dict with the information about the exchange transaction.
        """
        endpoint = ExchangeMoney
        path = endpoint.ROUTE
        body = endpoint.Body(
            from_=endpoint.Body.ModelFrom(
                account_id=from_account_id,
                currency=from_currency,
                amount=from_amount,
            ),
            to=endpoint.Body.ModelTo(
                account_id=to_account_id,
                currency=to_currency,
                amount=to_amount,
            ),
            reference=reference,
            request_id=request_id,
        )

        return await self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from pyrevolut.api.common import EnumTransactionState


class ResourceTransactionStateChanged(BaseModel):
    """
    Transaction state changed resource model.
    """

    id: Annotated[
        UUID,
        Field(description="The ID of the transaction."),
    ]
    request_id: Annotated[
        str | None,
        Field(description="The request ID provided by the client."),
    ] = None
    old_state: Annotated[
        EnumTransactionState,
        Field(description="The old state of the transaction."),
    ]
    new_state: Annotated[
        EnumTransactionState,
        Field(description="The new state of the transaction."),
    ]



from uuid import UUID
from datetime import datetime

from pyrevolut.utils import DateTime
from pyrevolut.api.common import (
    BaseEndpointSync,
    EnumTransactionType,
)

from pyrevolut.api.transactions.get import (
    RetrieveListOfTransactions,
    RetrieveTransaction,
)


class EndpointTransactionsSync(BaseEndpointSync):
    """The Transactions API

    Get the details of your transactions.

    Note
    ----
    An incoming or outgoing payment is represented as a transaction.
    """

    def get_all_transactions(
        self,
        from_datetime: datetime | DateTime | str | int | float | None = None,
        to_datetime: datetime | DateTime | str | int | float | None = None,
        account_id: UUID | None = None,
        limit: int | None = None,
        transaction_type: EnumTransactionType | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfTransactions.Response]:
        """
        Retrieve the historical transactions based on the provided query criteria.

        The transactions are sorted by the created_at date in reverse chronological order,
        and they're paginated. The maximum number of transactions returned per page is specified by the
        count parameter. To get the next page of results, make a new request and use the created_at date
        from the last item of the previous page as the value for the to parameter.

        Note
        ----
        The API returns a maximum of 1,000 transactions per request.

        Note
        ----
        To be compliant with PSD2 SCA regulations, businesses on the Revolut Business Freelancer
        plans can only access information older than 90 days within 5 minutes of the first authorisation.

        Parameters
        ----------
        from_datetime : datetime | DateTime | str | int | float, optional
            The date and time you retrieve the historical transactions from, including
            this date-time.
            Corresponds to the created_at value of the transaction.
            Provided in ISO 8601 format.

            Used also for pagination. To get back to the previous page of results,
            make a new request and use the created_at date from the first item of the
            current page as the value for the from parameter.
        to_datetime : datetime | DateTime | str | int | float, optional
            The date and time you retrieve the historical transactions to, excluding
            this date-time.
            Corresponds to the created_at value of the transaction.
            Provided in ISO 8601 format.
            The default value is the date and time at which you're calling the endpoint.

            Used also for pagination.
            To get the next page of results, make a new request and use the created_at
            date from the last item of the previous (current) page as the value for the
            to parameter.
        account_id : UUID, optional
            The ID of the account for which you want to retrieve the transactions.
        limit : int, optional
            The maximum number of transactions returned per page.
            To get the next page of results, make a new request and use the created_at
            date from the last item of the previous page as the value for the to parameter.
        transaction_type : EnumTransactionType, optional
            The type of the transaction.

        Returns
        -------
        list[dict] | list[RetrieveListOfTransactions.Response]
            A list of transactions.
        """
        endpoint = RetrieveListOfTransactions
        path = endpoint.ROUTE
        params = endpoint.Params(
            from_=from_datetime,
            to=to_datetime,
            account=account_id,
            count=limit,
            type=transaction_type,
        )

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_transaction(
        self,
        transaction_id: UUID | None = None,
        request_id: str | None = None,
        **kwargs,
    ) -> dict | RetrieveTransaction.Response:
        """
        Retrieve the details of a specific transaction.
        The details can include, for example, cardholder details for card payments.

        You can retrieve a transaction with its details either by its transaction ID
        or by the request ID that was provided for this transaction at the time of its
        creation, for example, when you created a payment.

        To retrieve a transaction by its transaction ID, use:

            /transaction/{transaction_id}

        To retrieve a transaction by a request ID provided at transaction creation, use:

            /transaction/{request_id}?id_type=request_id

        Parameters
        ----------
        transaction_id : UUID, optional
            The ID of the transaction.
            Specify either transaction_id or request_id.
        request_id : str, optional
            The request ID of the transaction.
            Specify either transaction_id or request_id.

        Returns
        -------
        dict | RetrieveTransaction.Response
            The details of the transaction.
        """
        assert (
            transaction_id or request_id
        ), "Either transaction_id or request_id must be provided."
        assert not (
            transaction_id and request_id
        ), "Either transaction_id or request_id must be provided, not both."

        endpoint = RetrieveTransaction
        path = endpoint.ROUTE.format(id=transaction_id or request_id)
        params = endpoint.Params(id_type="request_id" if request_id else None)

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )



from uuid import UUID
from datetime import date

from pyrevolut.utils import Date
from pyrevolut.api.common import BaseEndpointSync

from pyrevolut.api.payment_drafts.get import (
    RetrieveAllPaymentDrafts,
    RetrievePaymentDraft,
)
from pyrevolut.api.payment_drafts.post import CreatePaymentDraft
from pyrevolut.api.payment_drafts.delete import DeletePaymentDraft


class EndpointPaymentDraftsSync(BaseEndpointSync):
    """The Payment Drafts API

    Create a payment draft to request an approval for a payment from a
    business owner or admin before the payment is executed.
    The business owner or admin must manually approve it in the
    Revolut Business User Interface.

    You can also retrieve one or all payment drafts, and delete a payment draft.
    """

    def get_all_payment_drafts(
        self,
        **kwargs,
    ) -> dict | RetrieveAllPaymentDrafts.Response:
        """
        Get a list of all the payment drafts that aren't processed.

        Parameters
        ----------
        None

        Returns
        -------
        dict | RetrieveAllPaymentDrafts.Response
            A dict with the information about the payment drafts.
        """
        endpoint = RetrieveAllPaymentDrafts
        path = endpoint.ROUTE
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_payment_draft(
        self,
        payment_draft_id: UUID,
        **kwargs,
    ) -> dict | RetrievePaymentDraft.Response:
        """
        Get the information about a specific payment draft by ID.

        Parameters
        ----------
        payment_draft_id : UUID
            The ID of the payment draft.

        Returns
        -------
        dict | RetrievePaymentDraft.Response
            A dict with the information about the payment draft.
        """
        endpoint = RetrievePaymentDraft
        path = endpoint.ROUTE.format(payment_draft_id=payment_draft_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def create_payment_draft(
        self,
        account_id: UUID,
        counterparty_ids: list[UUID] = [],
        counterparty_account_ids: list[UUID | None] = [],
        counterparty_card_ids: list[UUID | None] = [],
        amounts: list[float] = [],
        currencies: list[str] = [],
        references: list[str] = [],
        title: str | None = None,
        schedule_for: date | Date | str | None = None,
        **kwargs,
    ) -> dict | CreatePaymentDraft.Response:
        """
        Create a payment draft.

        Parameters
        ----------
        account_id : UUID
            The ID of the account to pay from.
        counterparty_ids : list[UUID]
            The IDs of the counterparty accounts. Each ID corresponds to a payment.
        counterparty_account_ids : list[UUID | None]
            The IDs of the counterparty accounts. Each ID corresponds to a payment.
            If the counterparty has multiple payment methods available, use it to
            specify the account to which you want to send the money. Otherwise, use None.
        counterparty_card_ids : list[UUID | None]
            The IDs of the counterparty cards. Each ID corresponds to a payment.
            If the counterparty has multiple payment methods available, use it to
            specify the card to which you want to send the money. Otherwise, use None.
        amounts : list[float]
            The amounts of the payments.
        currencies : list[str]
            The ISO 4217 currency codes in upper case.
        references : list[str]
            The references for the payments.
        title : str, optional
            The title of the payment draft.
        schedule_for : date | Date | str, optional
            The scheduled date of the payment draft in ISO 8601 format.

        Returns
        -------
        dict | CreatePaymentDraft.Response
            A dict with the information about the payment draft created.
        """
        assert (
            len(counterparty_ids)
            == len(counterparty_account_ids)
            == len(counterparty_card_ids)
            == len(amounts)
            == len(currencies)
            == len(references)
        ), (
            "The number of elements in the lists must be equal. "
            f"Got {len(counterparty_ids)} counterparty_ids, "
            f"{len(counterparty_account_ids)} counterparty_account_ids, "
            f"{len(counterparty_card_ids)} counterparty_card_ids, "
            f"{len(amounts)} amounts, "
            f"{len(currencies)} currencies, "
            f"and {len(references)} references."
        )

        endpoint = CreatePaymentDraft
        path = endpoint.ROUTE
        body = endpoint.Body(
            title=title,
            schedule_for=schedule_for,
            payments=[
                endpoint.Body.ModelPayment(
                    account_id=account_id,
                    receiver=endpoint.Body.ModelPayment.ModelReceiver(
                        counterparty_id=counterparty_id,
                        account_id=counterparty_account_id,
                        card_id=counterparty_card_id,
                    ),
                    amount=amount,
                    currency=currency,
                    reference=reference,
                )
                for counterparty_id, counterparty_account_id, counterparty_card_id, amount, currency, reference in zip(
                    counterparty_ids,
                    counterparty_account_ids,
                    counterparty_card_ids,
                    amounts,
                    currencies,
                    references,
                )
            ],
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def delete_payment_draft(
        self,
        payment_draft_id: UUID,
        **kwargs,
    ) -> dict | DeletePaymentDraft.Response:
        """
        Delete a payment draft with the given ID.
        You can delete a payment draft only if it isn't processed.

        Parameters
        ----------
        payment_draft_id : UUID
            The ID of the payment draft.

        Returns
        -------
        dict | DeletePaymentDraft.Response
            A dict with the information about the payment draft deleted.
        """
        endpoint = DeletePaymentDraft
        path = endpoint.ROUTE.format(payment_draft_id=payment_draft_id)
        params = endpoint.Params()

        return self.client.delete(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )



from .common import PyRevolutBaseException


class PyRevolutBadRequest(PyRevolutBaseException):
    """Bad Request -- Your request is invalid."""

    def __init__(
        self,
        msg="Your request is invalid. Please check the request and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from enum import StrEnum


class EnumMerchantCategory(StrEnum):
    """Merchant category enumeration"""

    HEALTH = "health"
    GENERAL = "general"
    SERVICES = "services"
    AIRLINES = "airlines"
    TRANSPORT = "transport"
    ACCOMMODATION = "accommodation"
    UTILITIES = "utilities"
    SHOPPING = "shopping"
    FINANCIAL = "financial"
    FURNITURE = "furniture"
    HARDWARE = "hardware"
    GROCERIES = "groceries"
    FUEL = "fuel"
    ENTERTAINMENT = "entertainment"
    SOFTWARE = "software"
    RESTAURANTS = "restaurants"
    ADVERTISING = "advertising"
    CASH = "cash"
    EDUCATION = "education"
    GOVERNMENT = "government"



from enum import StrEnum


class EnumPayoutLinkState(StrEnum):
    """Payout link state.

    The state that the payout link is in. Possible states are:

    created:
        The payout link has been created, but the amount has not yet been blocked.
    failed:
        The payout link couldn't be generated due to a failure during transaction booking.
    awaiting:
        The payout link is awaiting approval.
    active:
        The payout link can be redeemed.
    expired:
        The payout link cannot be redeemed because it wasn't claimed before its expiry date.
    cancelled:
        The payout link cannot be redeemed because it was cancelled.
    processing:
        The payout link has been redeemed and is being processed.
    processed:
        The payout link has been redeemed and the money has been transferred to the recipient.
    """

    CREATED = "created"
    FAILED = "failed"
    AWAITING = "awaiting"
    ACTIVE = "active"
    EXPIRED = "expired"
    CANCELLED = "cancelled"
    PROCESSING = "processing"
    PROCESSED = "processed"



from .common import PyRevolutBaseException


class PyRevolutInternalServerError(PyRevolutBaseException):
    """Internal Server Error -- We had a problem with our server. Try again later."""

    def __init__(
        self,
        msg="The Revolut server had a problem. Please try again later.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from pydantic import BaseModel


class UnfreezeCard:
    """
    Unfreeze a card to re-enable spending for that card.
    You can only unfreeze a card that is in the state frozen.

    A successful unfreeze changes the card's state back to active,
    and no content is returned in the response.
    """

    ROUTE = "/1.0/cards/{card_id}/unfreeze"

    class Body(BaseModel):
        """
        Request body for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        pass



from pydantic import BaseModel


class TerminateCard:
    """
    Terminate a specific card, based on its ID.

    Once the card is terminated, it will not be returned by the API.

    A successful response does not get any content in return.
    """

    ROUTE = "cards/{card_id}"

    class Params(BaseModel):
        """Query parameters for the endpoint."""

        pass

    class Response(BaseModel):
        """Response model for the endpoint."""

        pass



"""This module holds all GET transactions endpoints."""

# flake8: noqa: F401
from .retrieve_list_of_transactions import RetrieveListOfTransactions
from .retrieve_transaction import RetrieveTransaction



from typing import Literal, Type
from uuid import UUID
from datetime import datetime

from pydantic import BaseModel

from pyrevolut.utils import DateTime
from pyrevolut.exceptions import PyRevolutInvalidEnvironment
from pyrevolut.api.common import BaseEndpointSync, EnumMerchantCategory
from pyrevolut.api.cards.get import (
    RetrieveListOfCards,
    RetrieveCardDetails,
    RetrieveSensitiveCardDetails,
)
from pyrevolut.api.cards.post import CreateCard, FreezeCard, UnfreezeCard
from pyrevolut.api.cards.patch import UpdateCardDetails
from pyrevolut.api.cards.delete import TerminateCard


class EndpointCardsSync(BaseEndpointSync):
    """The Cards API
    Manage cards for the business team members, freeze, unfreeze,
    terminate and update card settings, such as transaction limits.

    This feature is available in the UK, US and the EEA.
    This feature is not available in Sandbox.
    """

    def get_all_cards(
        self,
        created_before: datetime | DateTime | str | int | float | None = None,
        limit: int | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfCards.Response]:
        """
        Get the list of all cards in your organisation.
        The results are paginated and sorted by the created_at date in reverse chronological order.

        Parameters
        ----------
        created_before : datetime | DateTime | str | int | float | None
            Retrieves cards with created_at < created_before.
            The default value is the current date and time at which you are calling the endpoint.
            Provided in ISO 8601 format.
        limit : int | None
            The maximum number of cards returned per page.
            To get to the next page, make a new request and use the
            created_at date of the last card returned in the previous
            response as the value for created_before.

            If not provided, the default value is 100.

        Returns
        -------
        list[dict] | list[RetrieveListOfCards.Response]
            The list of all cards in your organisation.
        """
        self.__check_sandbox()
        endpoint = RetrieveListOfCards
        path = endpoint.ROUTE
        params = endpoint.Params(
            created_before=created_before,
            limit=limit,
        )

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_card(
        self,
        card_id: UUID,
        **kwargs,
    ) -> dict | RetrieveCardDetails.Response:
        """
        Get the details of a specific card, based on its ID.

        Parameters
        ----------
        card_id : UUID
            The card ID.

        Returns
        -------
        dict | RetrieveCardDetails.Response
            The details of the card.
        """
        self.__check_sandbox()
        endpoint = RetrieveCardDetails
        path = endpoint.ROUTE.format(card_id=card_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_card_sensitive_details(
        self,
        card_id: UUID,
        **kwargs,
    ) -> dict | RetrieveSensitiveCardDetails.Response:
        """
        Get sensitive details of a specific card, based on its ID.
        Requires the READ_SENSITIVE_CARD_DATA token scope.

        Parameters
        ----------
        card_id : UUID
            The card ID.

        Returns
        -------
        dict | RetrieveSensitiveCardDetails.Response
            The sensitive details of the card.
        """
        self.__check_sandbox()
        endpoint = RetrieveSensitiveCardDetails
        path = endpoint.ROUTE.format(card_id=card_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def create_card(
        self,
        request_id: str,
        holder_id: UUID,
        label: str | None = None,
        accounts: list[UUID] | None = None,
        categories: list[EnumMerchantCategory] | None = None,
        single_limit_amount: float | None = None,
        single_limit_currency: str | None = None,
        day_limit_amount: float | None = None,
        day_limit_currency: str | None = None,
        week_limit_amount: float | None = None,
        week_limit_currency: str | None = None,
        month_limit_amount: float | None = None,
        month_limit_currency: str | None = None,
        quarter_limit_amount: float | None = None,
        quarter_limit_currency: str | None = None,
        year_limit_amount: float | None = None,
        year_limit_currency: str | None = None,
        all_time_limit_amount: float | None = None,
        all_time_limit_currency: str | None = None,
        **kwargs,
    ) -> dict | CreateCard.Response:
        """
        Create a new card for an existing member of your Revolut Business team.

        When using the API, you can create only virtual cards.
        To create a physical card, use the Revolut Business app.

        Parameters
        ----------
        request_id : str
            A unique ID of the request that you provide.
            This ID is used to prevent duplicate card creation requests in case
            of a lost connection or client error, so make sure you use the same
            request_id for requests related to the same card.
            The deduplication is limited to 24 hours counting from the first request
            using a given ID.
        holder_id : UUID
            The ID of the team member who will be the holder of the card.
        label : str | None
            The label for the issued card, displayed in the UI to help distinguish between cards.
            If not specified, no label will be added.
        accounts : list[UUID] | None
            The list of accounts to link to the card. If not specified, all accounts will be linked.
        categories : list[EnumMerchantCategory] | None
            The list of merchant categories to link to the card. If not specified, all categories will be linked.
        single_limit_amount : float | None
            The maximum amount for a single transaction.
        single_limit_currency : str | None
            The currency of the single transaction limit.
        day_limit_amount : float | None
            The maximum amount for transactions in a day.
        day_limit_currency : str | None
            The currency of the day limit.
        week_limit_amount : float | None
            The maximum amount for transactions in a week.
        week_limit_currency : str | None
            The currency of the week limit.
        month_limit_amount : float | None
            The maximum amount for transactions in a month.
        month_limit_currency : str | None
            The currency of the month limit.
        quarter_limit_amount : float | None
            The maximum amount for transactions in a quarter.
        quarter_limit_currency : str | None
            The currency of the quarter limit.
        year_limit_amount : float | None
            The maximum amount for transactions in a year.
        year_limit_currency : str | None
            The currency of the year limit.
        all_time_limit_amount : float | None
            The maximum amount for transactions in the card's lifetime.
        all_time_limit_currency : str | None
            The currency of the all-time limit.

        Returns
        -------
        dict | CreateCard.Response
            The details of the created card.
        """
        self.__check_sandbox()
        endpoint = CreateCard
        path = endpoint.ROUTE

        # Create the SpendingLimits model (if applicable)
        spending_limits = endpoint.Body.ModelSpendingLimits(
            single=(
                endpoint.Body.ModelSpendingLimits.ModelSingle(
                    amount=single_limit_amount,
                    currency=single_limit_currency,
                )
                if single_limit_amount is not None and single_limit_currency is not None
                else None
            ),
            day=(
                endpoint.Body.ModelSpendingLimits.ModelDay(
                    amount=day_limit_amount,
                    currency=day_limit_currency,
                )
                if day_limit_amount is not None and day_limit_currency is not None
                else None
            ),
            week=(
                endpoint.Body.ModelSpendingLimits.ModelWeek(
                    amount=week_limit_amount,
                    currency=week_limit_currency,
                )
                if week_limit_amount is not None and week_limit_currency is not None
                else None
            ),
            month=(
                endpoint.Body.ModelSpendingLimits.ModelMonth(
                    amount=month_limit_amount,
                    currency=month_limit_currency,
                )
                if month_limit_amount is not None and month_limit_currency is not None
                else None
            ),
            quarter=(
                endpoint.Body.ModelSpendingLimits.ModelQuarter(
                    amount=quarter_limit_amount,
                    currency=quarter_limit_currency,
                )
                if quarter_limit_amount is not None
                and quarter_limit_currency is not None
                else None
            ),
            year=(
                endpoint.Body.ModelSpendingLimits.ModelYear(
                    amount=year_limit_amount,
                    currency=year_limit_currency,
                )
                if year_limit_amount is not None and year_limit_currency is not None
                else None
            ),
            all_time=(
                endpoint.Body.ModelSpendingLimits.ModelAllTime(
                    amount=all_time_limit_amount,
                    currency=all_time_limit_currency,
                )
                if all_time_limit_amount is not None
                and all_time_limit_currency is not None
                else None
            ),
        )
        if not any(
            [
                spending_limits.single is not None,
                spending_limits.day is not None,
                spending_limits.week is not None,
                spending_limits.month is not None,
                spending_limits.quarter is not None,
                spending_limits.year is not None,
                spending_limits.all_time is not None,
            ]
        ):
            spending_limits = None

        body = endpoint.Body(
            request_id=request_id,
            virtual=True,
            holder_id=holder_id,
            label=label,
            accounts=accounts,
            categories=categories,
            spending_limits=spending_limits,
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def freeze_card(
        self,
        card_id: UUID,
        **kwargs,
    ) -> dict | FreezeCard.Response:
        """
        Freeze a card to make it temporarily unavailable for spending.
        You can only freeze a card that is in the state active.

        A successful freeze changes the card's state to frozen,
        and no content is returned in the response.

        Parameters
        ----------
        card_id : UUID
            The card ID.

        Returns
        -------
        dict | FreezeCard.Response
            An empty dictionary.
        """
        self.__check_sandbox()
        endpoint = FreezeCard
        path = endpoint.ROUTE.format(card_id=card_id)
        body = endpoint.Body()

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def unfreeze_card(
        self,
        card_id: UUID,
        **kwargs,
    ) -> dict | UnfreezeCard.Response:
        """
        Unfreeze a card to make it available for spending again.
        You can only unfreeze a card that is in the state frozen.

        A successful unfreeze changes the card's state to active,
        and no content is returned in the response.

        Parameters
        ----------
        card_id : UUID
            The card ID.

        Returns
        -------
        dict | UnfreezeCard.Response
            An empty dictionary.
        """
        self.__check_sandbox()
        endpoint = UnfreezeCard
        path = endpoint.ROUTE.format(card_id=card_id)
        body = endpoint.Body()

        self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def update_card(
        self,
        card_id: UUID,
        label: str | None = None,
        categories: list[EnumMerchantCategory] | Literal["null"] | None = None,
        single_limit_amount: float | Literal["null"] | None = None,
        single_limit_currency: str | Literal["null"] | None = None,
        day_limit_amount: float | Literal["null"] | None = None,
        day_limit_currency: str | Literal["null"] | None = None,
        week_limit_amount: float | Literal["null"] | None = None,
        week_limit_currency: str | Literal["null"] | None = None,
        month_limit_amount: float | Literal["null"] | None = None,
        month_limit_currency: str | Literal["null"] | None = None,
        quarter_limit_amount: float | Literal["null"] | None = None,
        quarter_limit_currency: str | Literal["null"] | None = None,
        year_limit_amount: float | Literal["null"] | None = None,
        year_limit_currency: str | Literal["null"] | None = None,
        all_time_limit_amount: float | Literal["null"] | None = None,
        all_time_limit_currency: str | Literal["null"] | None = None,
        **kwargs,
    ) -> dict | UpdateCardDetails.Response:
        """
        Update details of a specific card, based on its ID.
        Updating a spending limit does not reset the spending counter.

        Parameters
        ----------
        card_id : UUID
            The card ID.
        label : str | None
            The label of the card.
        categories : list[EnumMerchantCategory] | Literal["null"] | None
            The list of merchant categories to link to the card.
            If set to 'null', all categories will be linked.
        single_limit_amount : float | Literal["null"] | None
            The maximum amount for a single transaction.
            If set to 'null', the limit will be removed.
        single_limit_currency : str | Literal["null"] | None
            The currency of the single transaction limit.
            If set to 'null', the limit will be removed.
        day_limit_amount : float | Literal["null"] | None
            The maximum amount for transactions in a day.
            If set to 'null', the limit will be removed.
        day_limit_currency : str | Literal["null"] | None
            The currency of the day limit.
            If set to 'null', the limit will be removed.
        week_limit_amount : float | Literal["null"] | None
            The maximum amount for transactions in a week.
            If set to 'null', the limit will be removed.
        week_limit_currency : str | Literal["null"] | None
            The currency of the week limit.
            If set to 'null', the limit will be removed.
        month_limit_amount : float | Literal["null"] | None
            The maximum amount for transactions in a month.
            If set to 'null', the limit will be removed.
        month_limit_currency : str | Literal["null"] | None
            The currency of the month limit.
            If set to 'null', the limit will be removed.
        quarter_limit_amount : float | Literal["null"] | None
            The maximum amount for transactions in a quarter.
            If set to 'null', the limit will be removed.
        quarter_limit_currency : str | Literal["null"] | None
            The currency of the quarter limit.
            If set to 'null', the limit will be removed.
        year_limit_amount : float | Literal["null"] | None
            The maximum amount for transactions in a year.
            If set to 'null', the limit will be removed.
        year_limit_currency : str | Literal["null"] | None
            The currency of the year limit.
            If set to 'null', the limit will be removed.
        all_time_limit_amount : float | Literal["null"] | None
            The maximum amount for transactions in the card's lifetime.
            If set to 'null', the limit will be removed.
        all_time_limit_currency : str | Literal["null"] | None
            The currency of the all-time limit.
            If set to 'null', the limit will be removed.

        Returns
        -------
        dict | UpdateCardDetails.Response
            The updated details of the card.
        """
        self.__check_sandbox()
        endpoint = UpdateCardDetails
        path = endpoint.ROUTE.format(card_id=card_id)

        # Create the SpendingLimits model (if applicable)
        spending_limits = endpoint.Body.ModelSpendingLimits(
            single=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelSingle,
                amount=single_limit_amount,
                currency=single_limit_currency,
            ),
            day=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelDay,
                amount=day_limit_amount,
                currency=day_limit_currency,
            ),
            week=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelWeek,
                amount=week_limit_amount,
                currency=week_limit_currency,
            ),
            month=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelMonth,
                amount=month_limit_amount,
                currency=month_limit_currency,
            ),
            quarter=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelQuarter,
                amount=quarter_limit_amount,
                currency=quarter_limit_currency,
            ),
            year=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelYear,
                amount=year_limit_amount,
                currency=year_limit_currency,
            ),
            all_time=self.__process_limit_model(
                model=endpoint.Body.ModelSpendingLimits.ModelAllTime,
                amount=all_time_limit_amount,
                currency=all_time_limit_currency,
            ),
        )
        if not any(
            [
                spending_limits.single is not None,
                spending_limits.day is not None,
                spending_limits.week is not None,
                spending_limits.month is not None,
                spending_limits.quarter is not None,
                spending_limits.year is not None,
                spending_limits.all_time is not None,
            ]
        ):
            spending_limits = None
        elif all(
            [
                spending_limits.single == "null",
                spending_limits.day == "null",
                spending_limits.week == "null",
                spending_limits.month == "null",
                spending_limits.quarter == "null",
                spending_limits.year == "null",
                spending_limits.all_time == "null",
            ]
        ):
            spending_limits = "null"

        body = endpoint.Body(
            label=label,
            categories=categories,
            spending_limits=spending_limits,
        )

        return self.client.patch(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def delete_card(
        self,
        card_id: UUID,
        **kwargs,
    ) -> dict | TerminateCard.Response:
        """
        Terminate a specific card, based on its ID.

        Once the card is terminated, it will not be returned by the API.

        A successful response does not get any content in return.

        Parameters
        ----------
        card_id : UUID
            The card ID.

        Returns
        -------
        dict | TerminateCard.Response
            An empty dictionary.
        """
        self.__check_sandbox()
        endpoint = TerminateCard
        path = endpoint.ROUTE.format(card_id=card_id)
        params = endpoint.Params()

        return self.client.delete(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def __process_limit_model(
        self,
        model: Type[BaseModel],
        amount: float | None,
        currency: str | None,
    ):
        """
        Process the limit model.
        """
        if amount is not None and currency is not None:
            return model(
                amount=amount,
                currency=currency,
            )
        elif amount == "null" and currency == "null":
            return "null"
        return None

    def __check_sandbox(self):
        """
        Check if the sandbox is enabled.

        Raises
        ------
        PyRevolutInvalidEnvironment
            If the sandbox is enabled.
        """
        if self.client.sandbox:
            raise PyRevolutInvalidEnvironment(
                "This feature is not available in Sandbox."
            )



"""This module contains the exceptions that are raised by the pyrevolut package."""

# flake8: noqa: F401

from .common import PyRevolutBaseException
from .bad_request import PyRevolutBadRequest
from .conflict import PyRevolutConflict
from .forbidden import PyRevolutForbidden
from .internal_server_error import PyRevolutInternalServerError
from .invalid_environment import PyRevolutInvalidEnvironment
from .invalid_payload import PyRevolutInvalidPayload
from .method_not_allowed import PyRevolutMethodNotAllowed
from .network_error import PyRevolutNetworkError
from .not_acceptable import PyRevolutNotAcceptable
from .not_found import PyRevolutNotFound
from .server_unavailable import PyRevolutServerUnavailable
from .timeout_error import PyRevolutTimeoutError
from .too_many_requests import PyRevolutTooManyRequests
from .unauthorized import PyRevolutUnauthorized



from pydantic import BaseModel


class DeleteWebhook:
    """
    Delete a specific webhook.

    A successful response does not get any content in return.
    """

    ROUTE = "/2.0/webhooks/{webhook_id}"

    class Params(BaseModel):
        """
        The query params of the request.
        """

        pass

    class Response(BaseModel):
        """
        The response model for the request.
        """

        pass



from typing import Type, Any
import datetime

import pendulum
from pendulum import Date as _Date  # type: ignore

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class Date(_Date):
    """A `pendulum.Date` object for Pydantic validation."""

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Return a Pydantic CoreSchema with the Date validation

        Parameters
        ----------
        source : Type[Any]
            The source type to be converted.
        handler : GetCoreSchemaHandler
            The handler to get the CoreSchema.

        Returns
        -------
        core_schema.CoreSchema
            A Pydantic CoreSchema with the Date validation.
        """
        return core_schema.no_info_wrap_validator_function(
            cls._validate, core_schema.date_schema()
        )

    @classmethod
    def _validate(
        cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler
    ) -> Any:
        """
        Validate the date object and return it.

        Parameters
        ----------
        value: Any
            The value to validate.
        handler: core_schema.ValidatorFunctionWrapHandler
            The handler to get the CoreSchema.

        Returns
        -------
        Any
            The validated value or raises a PydanticCustomError.
        """
        # if we are passed an existing instance, pass it straight through.
        if isinstance(value, _Date):
            return handler(value)

        # otherwise, parse it.
        try:
            data = to_date(value)
        except Exception as exc:
            raise PydanticCustomError("value_error", str(exc)) from exc  # type: ignore
        return handler(data)


def to_date(dt: datetime.date | _Date | str) -> Date:
    """Converts various date representations to a pendulum Date object.

    Parameters
    ----------
    dt : datetime.date | Date | str
        The object to be converted.

    Returns
    -------
    Date
        The pendulum Date object.
    """
    if isinstance(dt, datetime.date):
        return pendulum.Date(year=dt.year, month=dt.month, day=dt.day)
    elif isinstance(dt, str):
        return string_to_date(dt)
    else:
        raise ValueError(
            f"Unsupported type for conversion to pendulum Date: {type(dt)}"
        )


def string_to_date(string: str) -> Date:
    """Converts a string to a pendulum Date object.

    Parameters
    ----------
    string : str
        The string to be converted.

    Returns
    -------
    Date
        The pendulum Date object.
    """
    formats = [
        "YYYY-MM-DD",
        "YYYY/MM/DD",
        "YYYY.MM.DD",
        "YYYYMMDD",
        "YYYY-MM",
        "YYYY/MM",
        "YYYY.MM",
        "YYYYMM",
        "YYYY",
        "MM-YYYY",
        "MM/YYYY",
        "MM.YYYY",
        "MMYYYY",
    ]
    for format in formats:
        try:
            dt = pendulum.from_format(string=string, fmt=format, tz="UTC")
            return pendulum.Date(year=dt.year, month=dt.month, day=dt.day)
        except Exception:
            pass
    raise PydanticCustomError(f"Error converting string to pendulum Date: {string}")



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency
from pydantic_extra_types.country import CountryAlpha2, CountryAlpha3
from pydantic_extra_types.phone_numbers import PhoneNumber

from pyrevolut.utils import DateTime, Date
from pyrevolut.api.common import (
    EnumTransactionType,
    EnumTransactionState,
    EnumAccountType,
)


class ResourceTransaction(BaseModel):
    """
    Transaction resource model.
    """

    class ModelMerchant(BaseModel):
        """The information about the merchant (only for card transfers)"""

        name: Annotated[str, Field(description="The name of the merchant.")]
        city: Annotated[str, Field(description="The city of the merchant.")]
        category_code: Annotated[
            str, Field(description="The category code of the merchant.")
        ]
        country: Annotated[
            CountryAlpha2 | CountryAlpha3,
            Field(
                description="The country of the merchant as the 2-letter ISO 3166 code."
            ),
        ]

    class ModelLeg(BaseModel):
        """
        The legs of the transaction:

            - For transactions between your Revolut accounts,
            there can be 2 legs, for example, an internal transfer made out of
            the GBP account and into the EUR account.

            - For transactions in other cases, there is only 1 leg.
        """

        class ModelCounterparty(BaseModel):
            """The counterparty of the transaction."""

            account_id: Annotated[
                UUID | None,
                Field(description="The ID of the counterparty account."),
            ] = None
            account_type: Annotated[
                EnumAccountType,
                Field(description="Indicates the type of the account."),
            ]
            id: Annotated[
                UUID | None,
                Field(description="The ID of the counterparty."),
            ] = None

        leg_id: Annotated[UUID, Field(description="The ID of the leg.")]
        amount: Annotated[
            float,
            Field(description="The amount of the transaction leg."),
        ]
        fee: Annotated[
            float | None,
            Field(description="The amount of the transaction leg fee."),
        ] = None
        currency: Annotated[
            Currency,
            Field(description="ISO 4217 currency code in upper case."),
        ]
        bill_amount: Annotated[
            float | None,
            Field(description="The billing amount for cross-currency payments."),
        ] = None
        bill_currency: Annotated[
            Currency | None,
            Field(description="The billing currency for cross-currency payments."),
        ] = None
        account_id: Annotated[
            UUID,
            Field(
                description="The ID of the account that the transaction leg is associated with."
            ),
        ]
        counterparty: Annotated[
            ModelCounterparty | None,
            Field(description="The counterparty of the transaction leg."),
        ] = None
        description: Annotated[
            str | None,
            Field(description="The transaction leg purpose."),
        ] = None
        balance: Annotated[
            float | None,
            Field(
                description="The total balance of the account that the transaction is associated with."
            ),
        ] = None

    class ModelCard(BaseModel):
        """The card details (only for card transfers)."""

        card_number: Annotated[
            str,
            Field(description="The masked card number."),
        ]
        first_name: Annotated[
            str | None,
            Field(description="The first name of the cardholder."),
        ] = None
        last_name: Annotated[
            str | None,
            Field(description="The last name of the cardholder."),
        ] = None
        phone: Annotated[
            PhoneNumber | None,
            Field(
                description="The phone number of the cardholder in the E.164 format."
            ),
        ] = None

    id: Annotated[UUID, Field(description="The ID of the transaction.")]
    type: Annotated[
        EnumTransactionType, Field(description="Indicates the transaction type.")
    ]
    request_id: Annotated[
        str | None,
        Field(
            description="The request ID that you provided previously.",
        ),
    ] = None
    state: Annotated[
        EnumTransactionState, Field(description="Indicates the transaction state.")
    ]
    reason_code: Annotated[
        str | None,
        Field(
            description="The reason code when the transaction state is declined or failed."
        ),
    ] = None
    created_at: Annotated[
        DateTime,
        Field(
            description="The date and time the transaction was created in ISO 8601 format."
        ),
    ]
    updated_at: Annotated[
        DateTime,
        Field(
            description="The date and time the transaction was last updated in ISO 8601 format."
        ),
    ]
    completed_at: Annotated[
        DateTime | None,
        Field(
            description="""
            The date and time the transaction was completed in ISO 8601 format. 
            This is required when the transaction state is completed.
            """
        ),
    ] = None
    scheduled_for: Annotated[
        Date | None,
        Field(
            description="""
            The scheduled date of the payment, if applicable. Provided in ISO 8601 format.
            """
        ),
    ] = None
    related_transaction_id: Annotated[
        UUID | None,
        Field(description="The ID of the original transaction that has been refunded."),
    ] = None
    merchant: Annotated[
        ModelMerchant | None,
        Field(
            description="The information about the merchant (only for card transfers)."
        ),
    ] = None
    reference: Annotated[
        str | None,
        Field(description="The reference of the transaction."),
    ] = None
    legs: Annotated[
        list[ModelLeg],
        Field(
            description="""
            The legs of the transaction:

            - For transactions between your Revolut accounts,
            there can be 2 legs, for example, an internal transfer made out of
            the GBP account and into the EUR account.

            - For transactions in other cases, there is only 1 leg.
            """
        ),
    ]
    card: Annotated[
        ModelCard | None,
        Field(description="The card details (only for card transfers)."),
    ] = None



"""Use payout links to send money without having to request full banking details of the recipient.
The recipient must claim the money before the link expires.

Reference link: https://developer.revolut.com/docs/business/payout-links
"""

# flake8: noqa: F401
from .endpoint import EndpointPayoutLinksSync, EndpointPayoutLinksAsync



from uuid import UUID
from datetime import datetime

from pyrevolut.exceptions import PyRevolutInvalidEnvironment
from pyrevolut.utils.datetime import DateTime
from pyrevolut.api.common import BaseEndpointAsync
from pyrevolut.api.team_members.get import RetrieveListOfTeamMembers, RetrieveTeamRoles
from pyrevolut.api.team_members.post import InviteTeamMember


class EndpointTeamMembersAsync(BaseEndpointAsync):
    """The async Team Members API

    Retrieve information on existing team members of your organisation and invite new members.

    This feature is available in the UK, US and the EEA.

    This feature is not available in Sandbox.
    """

    async def get_team_members(
        self,
        created_before: datetime | DateTime | str | int | float | None = None,
        limit: int | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfTeamMembers.Response]:
        """
        Get information about all the team members of your business.

        The results are paginated and sorted by the created_at date in reverse chronological order.

        Note
        ----
        This feature is available in the UK, US and the EEA.

        This feature is not available in Sandbox.

        Parameters
        ----------
        created_before : datetime | DateTime | str | int | float | None
            Retrieves team members with created_at < created_before.
            The default value is the current date and time at which you are calling the endpoint.
            Provided in ISO 8601 format.
        limit : int | None
            The maximum number of team members returned per page.
            To get to the next page, make a new request and use the
            created_at date of the last team member returned in the previous
            response as the value for created_before.

            If not provided, the default value is 100.

        Returns
        -------
        list[dict] | list[RetrieveListOfTeamMembers.Response]
            The list of all team members in your organisation.
        """
        self.__check_sandbox()
        endpoint = RetrieveListOfTeamMembers
        path = endpoint.ROUTE
        params = endpoint.Params(
            created_before=created_before,
            limit=limit,
        )

        return await self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    async def get_team_roles(
        self,
        created_before: datetime | DateTime | str | int | float | None = None,
        limit: int | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveTeamRoles.Response]:
        """
        Get the list of roles for your business.

        The results are paginated and sorted by the created_at date in reverse chronological order.

        This feature is available in the UK, US and the EEA.

        This feature is not available in Sandbox.

        Parameters
        ----------
        created_before : datetime | DateTime | str | int | float | None
            Retrieves team roles with created_at < created_before.
            The default value is the current date and time at which you are calling the endpoint.
            Provided in ISO 8601 format.
        limit : int | None
            The maximum number of team roles returned per page.
            To get to the next page, make a new request and use the
            created_at date of the last role returned in the previous
            response as the value for created_before.

            If not provided, the default value is 100.

        Returns
        -------
        list[dict] | list[RetrieveTeamRoles.Response]
            The list of all team roles in your organisation.
        """
        self.__check_sandbox()
        endpoint = RetrieveTeamRoles
        path = endpoint.ROUTE
        params = endpoint.Params(
            created_before=created_before,
            limit=limit,
        )

        return await self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    async def invite_team_member(
        self,
        email: str,
        role_id: UUID | str,
        **kwargs,
    ) -> dict | InviteTeamMember.Response:
        """
        Invite a new member to your business account.

        When you invite a new team member to your business account,
        an invitation is sent to their email address that you provided in this request.
        To join your business account, the new team member has to accept this invitation.

        Note
        ----
        This feature is available in the UK, US and the EEA.

        This feature is not available in Sandbox.

        Parameters
        ----------
        email : str
            The email address of the invited member.
        role_id : UUID | str
            The ID of the role to assign to the new member.

        Returns
        -------
        dict | InviteTeamMember.Response
            The response model.
        """
        self.__check_sandbox()
        endpoint = InviteTeamMember
        path = endpoint.ROUTE
        body = endpoint.Body(
            email=email,
            role_id=role_id,
        )

        return await self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def __check_sandbox(self):
        """
        Check if the sandbox is enabled.

        Raises
        ------
        PyRevolutInvalidEnvironment
            If the sandbox is enabled.
        """
        if self.client.sandbox:
            raise PyRevolutInvalidEnvironment(
                "This feature is not available in Sandbox."
            )



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, EmailStr

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumTeamMemberState


class RetrieveListOfTeamMembers:
    """
    Get information about all the team members of your business.

    The results are paginated and sorted by the created_at date in reverse chronological order.

    Note
    ----
    This feature is available in the UK, US and the EEA.

    This feature is not available in Sandbox.
    """

    ROUTE = "/1.0/team-members"

    class Params(BaseModel):
        """The parameters of the request."""

        created_before: Annotated[
            DateTime | None,
            Field(
                description="""
                Retrieves team members with created_at < created_before. 
                The default value is the current date and time at which you are calling the endpoint.
                Provided in ISO 8601 format.
                """
            ),
        ] = None
        limit: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of team members returned per page.
                To get to the next page, make a new request and use the 
                created_at date of the last team member returned in the previous 
                response as the value for created_before.              
                
                If not provided, the default value is 100.  
                """,
                ge=1,
                le=100,
            ),
        ] = None

    class Response(BaseModel):
        """The response model."""

        id: Annotated[UUID, Field(description="The ID of the team member.")]
        email: Annotated[
            EmailStr, Field(description="The email address of the team member.")
        ]
        first_name: Annotated[
            str | None, Field(description="The team member's first name.")
        ] = None
        last_name: Annotated[
            str | None, Field(description="The team member's last name.")
        ] = None
        state: Annotated[
            EnumTeamMemberState,
            Field(description="The state that the team member is in."),
        ]
        role_id: Annotated[
            UUID | str,
            Field(
                description="The ID of the team member's role. This can be a UUID or other default role such as Owner."
            ),
        ]
        created_at: Annotated[
            DateTime,
            Field(
                description="The date and time the team member was created in ISO 8601 format."
            ),
        ]
        updated_at: Annotated[
            DateTime,
            Field(
                description="The date and time the team member was last updated in ISO 8601 format."
            ),
        ]



from datetime import datetime
import pendulum

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from pyrevolut.utils.datetime import DateTime, to_datetime


def gen_public_private_cert(
    expiration_dt: datetime | DateTime | str | int | float,
    country: str,
    email_address: str | None = None,
    common_name: str | None = None,
    state: str | None = None,
    locality: str | None = None,
    organization: str | None = None,
    organization_unit: str | None = None,
    save_location_public: str | None = "publiccert.cer",
    save_location_private: str | None = "privatecert.pem",
):
    """
    Method to generate a X509 RSA key pair and save it to a file.
    The key pair is used to create a client-assertion JWT for Revolut.

    Parameters
    ----------
    expiration_dt : datetime | DateTime | str | int | float
        The expiration datetime (UTC) of the certificate.
        This can be a datetime object, a DateTime object, a string, an integer, or a float.
    country : str
        The country of the certificate (2-letter code)
    email_address : str | None, optional
        The email address of the certificate.
    common_name : str | None, optional
        The common name (eg, fully qualified host name) of the certificate.
    state : str | None, optional
        The state (full name) of the certificate.
    locality : str | None, optional
        The locality (eg, city) of the certificate.
    organization : str | None, optional
        The organization (eg, company) of the certificate.
    organization_unit : str | None, optional
        The organization unit (eg, section) of the certificate.
    save_location_public : str | None, optional
        The location to save the public key.
        If None, the public key will not be saved.
        Default is "publiccert.cer".
    save_location_private : str | None, optional
        The location to save the private key.
        If None, the private key will not be saved.
        Default is "privatecert.pem".

    Returns
    -------
    dict
        A dictionary containing the public and private keys (in bytes)
    """
    # Generate an RSA key pair
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    # Create a certificate
    names = [x509.NameAttribute(NameOID.COUNTRY_NAME, country)]
    if email_address is not None:
        names.append(x509.NameAttribute(NameOID.EMAIL_ADDRESS, email_address))
    if common_name is not None:
        names.append(x509.NameAttribute(NameOID.COMMON_NAME, common_name))
    if state is not None:
        names.append(x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state))
    if locality is not None:
        names.append(x509.NameAttribute(NameOID.LOCALITY_NAME, locality))
    if organization is not None:
        names.append(x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization))
    if organization_unit is not None:
        names.append(
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, organization_unit)
        )
    subject = issuer = x509.Name(names)

    # Create a self-signed certificate
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(pendulum.now(tz="UTC"))
        .not_valid_after(to_datetime(dt=expiration_dt))
        .sign(private_key, hashes.SHA256())
    )

    # Get the public/private key in bytes
    public_key = cert.public_bytes(serialization.Encoding.PEM)
    private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    # Save the private key to a file if save_location is provided
    if save_location_private is not None:
        with open(save_location_private, "wb") as f:
            f.write(private_key)

    # Save the public key to a file if save_location is provided
    if save_location_public is not None:
        with open(save_location_public, "wb") as f:
            f.write(public_key)

    # Return the keys
    return {
        "public": public_key,
        "private": private_key,
    }



from enum import StrEnum


class EnumAuthScope(StrEnum):
    """Enum class for the scopes

    Access tokens can be issued with four security scopes and require a JWT (JSON Web Token)
    signature to be obtained:

        READ: Permissions for GET operations.

        WRITE: Permissions to update counterparties, webhooks, and issue payment drafts.

        PAY: Permissions to initiate or cancel transactions and currency exchanges.

        READ_SENSITIVE_CARD_DATA: Permissions to retrieve sensitive card details.

    Caution
    -------
    If you enable the READ_SENSITIVE_CARD_DATA scope for your access token, you must
    set up IP whitelisting.
    Failing to do so will prevent you from accessing any Business API endpoint.
    """

    READ = "READ"
    WRITE = "WRITE"
    PAY = "PAY"
    READ_SENSITIVE_CARD_DATA = "READ_SENSITIVE_CARD_DATA"



from enum import StrEnum


class EnumAccountNameMatchReasonCode(StrEnum):
    """Account Name Match Reason Code Enum

    The reason code. Possible values:

        close_match:
            The provided name is similar to the account name, the account type is correct.
            The actual name is returned.
        individual_account_name_matched:
            The names match but the counterparty is an individual, not a business.
        company_account_name_matched:
            The names match but the counterparty is a business, not an individual.
        individual_account_close_match:
            The provided name is similar to the account name, and the account type is incorrect
            - the counterparty is an individual, not a business. The actual name is returned.
        company_account_close_match:
            The provided name is similar to the account name, and the account type is incorrect
            - the counterparty is a business, not an individual. The actual name is returned.
        not_matched:
            The account details don't match the provided values.
        account_does_not_exist:
            The account does not exist.
        account_switched:
            The account has been switched using the Current Account Switching Service.
            Please contact the recipient for updated account details.
        cannot_be_checked:
            The account cannot be checked.
    """

    CLOSE_MATCH = "close_match"
    INDIVIDUAL_ACCOUNT_NAME_MATCHED = "individual_account_name_matched"
    COMPANY_ACCOUNT_NAME_MATCHED = "company_account_name_matched"
    INDIVIDUAL_ACCOUNT_CLOSE_MATCH = "individual_account_close_match"
    COMPANY_ACCOUNT_CLOSE_MATCH = "company_account_close_match"
    NOT_MATCHED = "not_matched"
    ACCOUNT_DOES_NOT_EXIST = "account_does_not_exist"
    ACCOUNT_SWITCHED = "account_switched"
    CANNOT_BE_CHECKED = "cannot_be_checked"



from enum import StrEnum


class EnumPaymentScheme(StrEnum):
    """Payment scheme enum"""

    CHAPS = "chaps"
    BACS = "bacs"
    FASTER_PAYMENTS = "faster_payments"
    SEPA = "sepa"
    SWIFT = "swift"
    ACH = "ach"
    ELIXIR = "elixir"
    SORBNET = "sorbnet"
    NICS = "nics"
    RIX = "rix"
    SUMCLEARING = "sumclearing"



"""Retrieve information on exchange rates between currencies, buy and sell currencies.

Reference link: https://developer.revolut.com/docs/business/foreign-exchange
"""

# flake8: noqa: F401
from .endpoint import EndpointForeignExchangeSync, EndpointForeignExchangeAsync



from typing import Annotated, Literal

from pydantic import BaseModel, Field

from pyrevolut.api.transactions.resources import ResourceTransaction


class RetrieveTransaction:
    """
    Retrieve the details of a specific transaction.
    The details can include, for example, cardholder details for card payments.

    You can retrieve a transaction with its details either by its transaction ID
    or by the request ID that was provided for this transaction at the time of its
    creation, for example, when you created a payment.

    To retrieve a transaction by its transaction ID, use:

        /transaction/{transaction_id}

    To retrieve a transaction by a request ID provided at transaction creation, use:

        /transaction/{request_id}?id_type=request_id
    """

    ROUTE = "/1.0/transaction/{id}"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        id_type: Annotated[
            Literal["request_id"] | None,
            Field(
                description="""
                The type of the ID you're providing for the transaction retrieval.
                The default value is None, which means that the transaction ID is provided.
                """,
            ),
        ] = None

    class Response(ResourceTransaction):
        """
        The response model for the request.
        """

        pass



from typing import Annotated
import json

import pendulum
from pydantic import BaseModel, Field, SecretStr, field_serializer, ConfigDict

from pyrevolut.utils.datetime import DateTime


class ModelCreds(BaseModel):
    """The model that represents the credentials JSON file."""

    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    class ModelCertificate(BaseModel):
        """The model that represents the certificate information"""

        model_config = ConfigDict(validate_assignment=True, extra="forbid")

        public: Annotated[
            SecretStr,
            Field(description="The public certificate in base64 encoded format"),
        ]
        private: Annotated[
            SecretStr,
            Field(description="The private certificate in base64 encoded format"),
        ]
        expiration_dt: Annotated[
            DateTime, Field(description="The expiration datetime of the certificates")
        ]

        @field_serializer("public", "private", when_used="json")
        def dump_secret(self, value: SecretStr) -> str:
            """Serialize the secret value to a string"""
            return value.get_secret_value()

    class ModelClientAssertJWT(BaseModel):
        """The model that represents the client assertion JWT information"""

        model_config = ConfigDict(validate_assignment=True, extra="forbid")

        jwt: Annotated[SecretStr, Field(description="The JWT assertion string")]
        expiration_dt: Annotated[
            DateTime, Field(description="The expiration datetime of the JWT")
        ]

        @field_serializer("jwt", when_used="json")
        def dump_secret(self, value: SecretStr) -> str:
            """Serialize the secret value to a string"""
            return value.get_secret_value()

    class ModelTokens(BaseModel):
        """The model that represents the tokens information"""

        model_config = ConfigDict(validate_assignment=True, extra="forbid")

        access_token: Annotated[SecretStr, Field(description="The access token")]
        refresh_token: Annotated[SecretStr, Field(description="The refresh token")]
        token_type: Annotated[str, Field(description="The token type")]
        access_token_expiration_dt: Annotated[
            DateTime, Field(description="The expiration datetime of the access token")
        ]
        refresh_token_expiration_dt: Annotated[
            DateTime, Field(description="The expiration datetime of the refresh token")
        ]

        @field_serializer("access_token", "refresh_token", when_used="json")
        def dump_secret(self, value: SecretStr) -> str:
            """Serialize the secret value to a string"""
            return value.get_secret_value()

    certificate: Annotated[
        ModelCertificate, Field(description="The certificate information")
    ]
    client_assert_jwt: Annotated[
        ModelClientAssertJWT, Field(description="The client assertion JWT information")
    ]
    tokens: Annotated[ModelTokens, Field(description="The tokens information")]

    @property
    def access_token_expired(self) -> bool:
        """Check if the access token has expired.
        This means that the token is no longer valid and a new one should be requested.

        Returns
        -------
        bool
        """
        # Subtract 1 minute to ensure that the token still works for a little bit
        return self.tokens.access_token_expiration_dt.subtract(
            minutes=1
        ) < pendulum.now(tz="UTC")

    @property
    def credentials_expired(self) -> bool:
        """Check if any of the long term credentials have expired.
        This means that the certificate, client assertion JWT, or refresh token has expired and
        a new one should be requested.

        Returns
        -------
        bool
            True if any of the credentials have expired, False otherwise.
        """
        dt_now = pendulum.now(tz="UTC")
        if self.certificate.expiration_dt < dt_now:
            return True
        if self.client_assert_jwt.expiration_dt < dt_now:
            return True
        if self.tokens.refresh_token_expiration_dt < dt_now:
            return True
        return False


def save_creds(
    creds: ModelCreds,
    location: str = "credentials.json",
    indent: int = 4,
):
    """Save the credentials to the provided location.

    Parameters
    ----------
    creds : ModelCreds
        The credentials model
    location : str, optional
        The location to save the credentials to, by default "credentials.json"
    indent : int, optional
        The indentation level to use, by default 4

    Returns
    -------
    None
    """
    with open(location, "w") as file:
        json.dump(creds.model_dump(mode="json"), file, indent=indent)


def load_creds(location: str = "credentials.json") -> ModelCreds:
    """Load the credentials from the provided location.

    Parameters
    ----------
    location : str, optional
        The location to load the credentials from, by default "credentials.json"

    Returns
    -------
    ModelCreds
        The credentials model
    """
    with open(location, "r") as file:
        return ModelCreds(**json.load(file))



from .common import PyRevolutBaseException


class PyRevolutNotAcceptable(PyRevolutBaseException):
    """Not Acceptable -- You requested a format that isn't JSON."""

    def __init__(
        self,
        msg="You requested a format that isn't JSON. Please check the request and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from typing import Type

from pydantic import BaseModel

from httpx import Client as HTTPClient

from pyrevolut.api import (
    EndpointAccountsSync,
    EndpointCardsSync,
    EndpointCounterpartiesSync,
    EndpointForeignExchangeSync,
    EndpointPaymentDraftsSync,
    EndpointPayoutLinksSync,
    EndpointSimulationsSync,
    EndpointTeamMembersSync,
    EndpointTransactionsSync,
    EndpointTransfersSync,
    EndpointWebhooksSync,
)

from .base import BaseClient


class Client(BaseClient):
    """The synchronous client for the Revolut API"""

    Accounts: EndpointAccountsSync
    Cards: EndpointCardsSync
    Counterparties: EndpointCounterpartiesSync
    ForeignExchange: EndpointForeignExchangeSync
    PaymentDrafts: EndpointPaymentDraftsSync
    PayoutLinks: EndpointPayoutLinksSync
    Simulations: EndpointSimulationsSync
    TeamMembers: EndpointTeamMembersSync
    Transactions: EndpointTransactionsSync
    Transfers: EndpointTransfersSync
    Webhooks: EndpointWebhooksSync

    def open(self):
        """Opens the client connection"""
        if self.client is not None:
            return

        self.client = HTTPClient()

    def close(self):
        """Closes the client connection"""
        if self.client is None:
            return

        self.client.close()
        self.client = None

    def get(
        self,
        path: str,
        response_model: Type[BaseModel],
        params: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send a GET request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        params : Type[BaseModel] | None
            The parameters to add to the request route

        Returns
        -------
        Response
            The response from the request
        """
        resp = self.client.get(
            **self._prep_get(
                path=path,
                params=params,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    def post(
        self,
        path: str,
        response_model: Type[BaseModel],
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send a POST request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        body : Type[BaseModel] | None
            The body to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = self.client.post(
            **self._prep_post(
                path=path,
                body=body,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    def patch(
        self,
        path: str,
        response_model: Type[BaseModel],
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send a PATCH request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        body : Type[BaseModel]
            The body to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = self.client.patch(
            **self._prep_patch(
                path=path,
                body=body,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    def delete(
        self,
        path: str,
        response_model: Type[BaseModel],
        params: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send a DELETE request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        params : Type[BaseModel] | None
            The parameters to add to the request route

        Returns
        -------
        Response
            The response from the request
        """
        resp = self.client.delete(
            **self._prep_delete(
                path=path,
                params=params,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    def put(
        self,
        response_model: Type[BaseModel],
        path: str,
        body: Type[BaseModel] | None = None,
        **kwargs,
    ):
        """Send a PUT request to the Revolut API

        Parameters
        ----------
        path : str
            The path to send the request to
        response_model : Type[BaseModel]
            The model to use for the response
        body : Type[BaseModel] | None
            The body to send in the request

        Returns
        -------
        Response
            The response from the request
        """
        resp = self.client.put(
            **self._prep_put(
                path=path,
                body=body,
                **kwargs,
            )
        )
        return self.process_response(
            response=resp,
            response_model=response_model,
            return_type=None,
            error_response=None,
        )

    def load_endpoints(self):
        """Loads all the endpoints from the api directory"""
        self.Accounts = EndpointAccountsSync(client=self)
        self.Cards = EndpointCardsSync(client=self)
        self.Counterparties = EndpointCounterpartiesSync(client=self)
        self.ForeignExchange = EndpointForeignExchangeSync(client=self)
        self.PaymentDrafts = EndpointPaymentDraftsSync(client=self)
        self.PayoutLinks = EndpointPayoutLinksSync(client=self)
        self.Simulations = EndpointSimulationsSync(client=self)
        self.TeamMembers = EndpointTeamMembersSync(client=self)
        self.Transactions = EndpointTransactionsSync(client=self)
        self.Transfers = EndpointTransfersSync(client=self)
        self.Webhooks = EndpointWebhooksSync(client=self)

    def __enter__(self):
        """Open the client connection"""
        self.open()
        return self

    def __exit__(self, *args, **kwargs):
        """Close the client connection"""
        self.close()



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency
from pydantic_extra_types.pendulum_dt import Duration

from pyrevolut.api.common import EnumTransferReasonCode, EnumPayoutLinkPaymentMethod
from pyrevolut.api.payout_links.resources import ResourcePayoutLink


class CreatePayoutLink:
    """
    Create a payout link to send money even when you don't have the full
    banking details of the counterparty.
    After you have created the link, send it to the recipient so that
    they can claim the payment.

    Note
    ----
    This feature is available in the UK and the EEA.
    """

    ROUTE = "/1.0/payout-links"

    class Body(BaseModel):
        """
        The request body model for the request.
        """

        counterparty_name: Annotated[
            str,
            Field(
                description="""
                The name of the counterparty provided by the sender.                
                """,
            ),
        ]
        save_counterparty: Annotated[
            bool | None,
            Field(
                description="""
                Indicates whether to save the recipient as your counterparty upon link claim. 
                If false then the counterparty will not show up on your counterparties list, 
                for example, when you retrieve your counterparties. 
                However, you will still be able to retrieve this counterparty by its ID.

                If you don't choose to save the counterparty on link creation, you can do it later 
                from your transactions list in the Business app.
                
                If not provided, the default value is false.
                """,
            ),
        ] = None
        request_id: Annotated[
            str,
            Field(
                description="""
                The ID of the request, provided by the sender.
                
                To ensure that a link payment is not processed multiple times if there
                are network or system errors, the same request_id should be used for 
                requests related to the same link.
                """,
                max_length=40,
            ),
        ]
        account_id: Annotated[
            UUID,
            Field(
                description="""
                The ID of the sender's account.
                """,
            ),
        ]
        amount: Annotated[
            float,
            Field(
                description="""
                The amount of money to be transferred.
                The amount must be between £1 and £2,500, or equivalent in the selected currency.
                """,
            ),
        ]
        currency: Annotated[
            Currency,
            Field(
                description="""
                ISO 4217 currency code in upper case.
                """,
            ),
        ]
        reference: Annotated[
            str,
            Field(
                description="""
                The reference for the payout transaction, provided by the sender.
                """,
            ),
        ]
        payout_methods: Annotated[
            list[EnumPayoutLinkPaymentMethod],
            Field(
                description="""
                The list of payout methods that the recipient can use to claim the payout, where:

                    revolut:
                        Revolut peer-to-peer (P2P) transfer

                    bank_account:
                        External bank transfer

                    card:
                        Card transfer
                """
            ),
        ]
        expiry_period: Annotated[
            Duration | None,
            Field(
                description="""
                Possible values: >= P1D and <= P7D

                Default value: P7D

                The period after which the payout link expires if not claimed before, 
                provided in ISO 8601 format.

                The default and maximum value is 7 days from the link creation.
                """,
            ),
        ] = None
        transfer_reason_code: Annotated[
            EnumTransferReasonCode | None,
            Field(
                description="""
                The reason code for the transaction. 
                Transactions to certain countries and currencies might require you to 
                provide a transfer reason. 
                You can check available reason codes with the getTransferReasons operation.

                If a transfer reason is not required for the given currency and country, 
                this field is ignored.
                """
            ),
        ] = None

    class Response(ResourcePayoutLink):
        """
        The response model for the request.
        """

        pass



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, HttpUrl

from pyrevolut.utils import DateTime
from pyrevolut.api.webhooks.resources import ResourceWebhookPayload


class RetrieveListOfFailedWebhooks:
    """
    Get the list of all your failed webhook events, or use the query
    parameters to filter the results.

    The events are sorted by the created_at date in reverse chronological order.

    The returned failed events are paginated. The maximum number of events returned
    per page is specified by the limit parameter.
    To get to the next page, make a new request and use the created_at date of the
    last event returned in the previous response.
    """

    ROUTE = "/2.0/webhooks/{webhook_id}/failed-events"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        limit: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of events returned per page.

                To get to the next page, make a new request and use the created_at date of 
                the last event returned in the previous response as value for created_before.
                
                If not specified, the default value is 100.
                """,
                ge=1,
                le=1000,
            ),
        ] = None
        created_before: Annotated[
            DateTime | None,
            Field(
                description="""
                Retrieves events with created_at < created_before. 
                Cannot be older than the current date minus 21 days. 
                The default value is the current date and time at which you are calling the endpoint.

                Provided in ISO 8601 format.
                """,
            ),
        ] = None

    class Response(BaseModel):
        """
        The response model for the request.
        """

        id: Annotated[UUID, Field(description="The ID of the webhook event.")]
        created_at: Annotated[
            DateTime,
            Field(
                description="The date and time the event was created in ISO 8601 format.",
            ),
        ]
        updated_at: Annotated[
            DateTime,
            Field(
                description="The date and time the event was last updated in ISO 8601 format.",
            ),
        ]
        webhook_id: Annotated[
            UUID,
            Field(description="The ID of the webhook for which the event failed."),
        ]
        webhook_url: Annotated[
            HttpUrl,
            Field(
                description="The valid webhook URL that event notifications are sent to. The supported protocol is https",
            ),
        ]
        payload: Annotated[
            ResourceWebhookPayload,
            Field(description="The details of the failed event."),
        ]
        last_sent_date: Annotated[
            DateTime,
            Field(
                description="The date and time the last attempt at the event delivery occurred in ISO 8601 format.",
            ),
        ]



from enum import StrEnum


class EnumAccountNameMatchReasonType(StrEnum):
    """Account Name Match Reason Type Enum

    The reason type. Possible values:

        uk_cop:
            The CoP reason.
    """

    UK_COP = "uk_cop"



from .common import PyRevolutBaseException


class PyRevolutConflict(PyRevolutBaseException):
    """Conflict -- Your request conflicts with current state of a resource."""

    def __init__(
        self,
        msg="Your request conflicts with the current state of a resource. Please check the request and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



"""This module holds all GET payment drafts endpoints."""

# flake8: noqa: F401
from .retrieve_all_payment_drafts import RetrieveAllPaymentDrafts
from .retrieve_payment_draft import RetrievePaymentDraft



from uuid import UUID
from datetime import datetime

from pydantic_extra_types.pendulum_dt import Duration
from pyrevolut.utils import DateTime
from pyrevolut.api.common import (
    BaseEndpointSync,
    EnumPayoutLinkState,
    EnumPayoutLinkPaymentMethod,
    EnumTransferReasonCode,
)

from pyrevolut.api.payout_links.get import RetrieveListOfPayoutLinks, RetrievePayoutLink
from pyrevolut.api.payout_links.post import CreatePayoutLink, CancelPayoutLink


class EndpointPayoutLinksSync(BaseEndpointSync):
    """The Payout Links API

    Use payout links to send money without having to request full
    banking details of the recipient.
    The recipient must claim the money before the link expires.
    """

    def get_all_payout_links(
        self,
        state: EnumPayoutLinkState | None = None,
        created_before: datetime | DateTime | str | int | float | None = None,
        limit: int | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfPayoutLinks.Response]:
        """
        Get all the links that you have created, or use the query parameters to filter the results.

        The links are sorted by the created_at date in reverse chronological order.

        The returned links are paginated. The maximum number of payout links returned per
        page is specified by the limit parameter. To get to the next page, make a
        new request and use the created_at date of the last payout link returned in the previous response.

        Note
        ----
        This feature is available in the UK and the EEA.

        Parameters
        ----------
        state : EnumPayoutLinkState, optional
            The state that the payout link is in. Possible states are:

                created:
                    The payout link has been created, but the amount has not yet been blocked.
                failed:
                    The payout link couldn't be generated due to a failure during transaction booking.
                awaiting:
                    The payout link is awaiting approval.
                active:
                    The payout link can be redeemed.
                expired:
                    The payout link cannot be redeemed because it wasn't claimed before its expiry date.
                cancelled:
                    The payout link cannot be redeemed because it was cancelled.
                processing:
                    The payout link has been redeemed and is being processed.
                processed:
                    The payout link has been redeemed and the money has been transferred to the recipient.
        created_before : datetime | DateTime | str | int | float, optional
            Retrieves links with created_at < created_before.
            The default value is the current date and time at which you are calling the endpoint.

            Provided in ISO 8601 format.
        limit : int, optional
            The maximum number of links returned per page.
            To get to the next page, make a new request and use the
            created_at date of the last payout link returned in the previous
            response as the value for created_before.

            If not provided, the default value is 100.

        Returns
        -------
        list[dict] | list[RetrieveListOfPayoutLinks.Response]
            A list of payout links.
        """
        endpoint = RetrieveListOfPayoutLinks
        path = endpoint.ROUTE
        params = endpoint.Params(
            state=state,
            created_before=created_before,
            limit=limit,
        )

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_payout_link(
        self,
        payout_link_id: UUID,
        **kwargs,
    ) -> dict | RetrievePayoutLink.Response:
        """
        Get the information about a specific link by its ID.

        Note
        ----
        This feature is available in the UK and the EEA.

        Parameters
        ----------
        payout_link_id : UUID
            The ID of the payout link.

        Returns
        -------
        dict | RetrievePayoutLink.Response
            The payout link information.
        """
        endpoint = RetrievePayoutLink
        path = endpoint.ROUTE.format(payout_link_id=payout_link_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def create_payout_link(
        self,
        counterparty_name: str,
        request_id: str,
        account_id: UUID,
        amount: float,
        currency: str,
        reference: str,
        payout_methods: list[EnumPayoutLinkPaymentMethod],
        save_counterparty: bool | None = None,
        expiry_period: Duration | str | None = None,
        transfer_reason_code: EnumTransferReasonCode | None = None,
        **kwargs,
    ) -> dict | CreatePayoutLink.Response:
        """
        Create a payout link to send money even when you don't have the full
        banking details of the counterparty.
        After you have created the link, send it to the recipient so that
        they can claim the payment.

        Note
        ----
        This feature is available in the UK and the EEA.

        Parameters
        ----------
        counterparty_name : str
            The name of the counterparty provided by the sender.
        request_id : str
            The ID of the request, provided by the sender.

            To ensure that a link payment is not processed multiple times if there
            are network or system errors, the same request_id should be used for
            requests related to the same link.
        account_id : UUID
            The ID of the sender's account.
        amount : float
            The amount of money to be sent.
        currency : str
            The currency of the amount to be sent.
        reference : str
            A reference for the payment.
        payout_methods : list[EnumPayoutLinkPaymentMethod]
            The payout methods that the recipient can use to claim the payment.
            If not provided, the default value is
            [EnumPayoutLinkPaymentMethod.REVOLUT, EnumPayoutLinkPaymentMethod.BANK_ACCOUNT].
        save_counterparty : bool, optional
            Indicates whether to save the recipient as your counterparty upon link claim.
            If false then the counterparty will not show up on your counterparties list,
            for example, when you retrieve your counterparties.
            However, you will still be able to retrieve this counterparty by its ID.

            If you don't choose to save the counterparty on link creation, you can do it later
            from your transactions list in the Business app.

            If not provided, the default value is false.
        expiry_period : Duration | str, optional
            Possible values: >= P1D and <= P7D

            Default value: P7D

            The period after which the payout link expires if not claimed before,
            provided in ISO 8601 format.

            The default and maximum value is 7 days from the link creation.
        transfer_reason_code : EnumTransferReasonCode, optional
            The reason code for the transaction.
            Transactions to certain countries and currencies might require you to
            provide a transfer reason.
            You can check available reason codes with the getTransferReasons operation.

            If a transfer reason is not required for the given currency and country,
            this field is ignored.

        Returns
        -------
        dict | CreatePayoutLink.Response
            The payout link information.
        """
        endpoint = CreatePayoutLink
        path = endpoint.ROUTE
        body = endpoint.Body(
            counterparty_name=counterparty_name,
            save_counterparty=save_counterparty,
            request_id=request_id,
            account_id=account_id,
            amount=amount,
            currency=currency,
            reference=reference,
            payout_methods=payout_methods,
            expiry_period=expiry_period,
            transfer_reasion_code=transfer_reason_code,
        )

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    def cancel_payout_link(
        self,
        payout_link_id: UUID,
        **kwargs,
    ) -> dict | CancelPayoutLink.Response:
        """
        Cancel a payout link.
        You can only cancel a link that hasn't been claimed yet.
        A successful request does not get any content in response.

        Note
        ----
        This feature is available in the UK and the EEA.

        Parameters
        ----------
        payout_link_id : UUID
            The ID of the payout link.

        Returns
        -------
        dict | CancelPayoutLink.Response
            An empty dictionary.
        """
        endpoint = CancelPayoutLink
        path = endpoint.ROUTE.format(payout_link_id=payout_link_id)
        body = endpoint.Body()

        return self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )



from .common import PyRevolutBaseException


class PyRevolutMethodNotAllowed(PyRevolutBaseException):
    """Method Not Allowed -- You tried to access an endpoint with an invalid method."""

    def __init__(
        self,
        msg="You tried to access an endpoint with an invalid method. Please check the method and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



import hmac
import hashlib
import json

import pendulum

from pyrevolut.api.common import BaseEndpointSync
from pyrevolut.exceptions import PyRevolutInvalidPayload
from pyrevolut.api.webhooks.resources import ResourceWebhookPayload


class BaseEndpointWebhooks(BaseEndpointSync):
    """The Webhooks API

    A webhook (also called a web callback) allows your system to receive
    updates about your account to an HTTPS endpoint that you provide.
    When a supported event occurs, a notification is posted via HTTP POST method
    to the specified endpoint.

    If the receiver returns an HTTP error response, Revolut will retry the webhook
    event three more times, each with a 10-minute interval.

    The following events are supported:

    TransactionCreated
    TransactionStateChanged
    PayoutLinkCreated
    PayoutLinkStateChanged
    """

    def receive_webhook_event(
        self,
        raw_payload: str,
        signing_secret: str,
        header_timestamp: int,
        header_signature: str,
        **kwargs,
    ):
        """
        Receive a webhook event notification. Will verify the payload signature and
        raise an exception if the signature is invalid.

        Parameters
        ----------
        raw_payload : str
            The raw payload string received from the webhook event.
        signing_secret : str
            The signing secret provided by Revolut for the webhook endpoint.
        header_timestamp : int
            The timestamp string received from the webhook event. It
            will be in the header of the request under the key `Revolut-Requested-Timestamp`.
            For example: 1683650202360
        header_signature : str
            Signature of the request payload. It will be in the header of the request
            under the key `Revolut-Signature`.
            Contains the current version of the signature generating algorithm,
            and the hexadecimal-encoded signature itself.
            For example: v1=09a9989dd8d9282c1d34974fc730f5cbfc4f4296941247e90ae5256590a11e8c.

            It can be that there are multiple signatures in the header if multiple signing secrets are active at a given moment.
            If that's the case, they are separated by a comma.
            For example:
            Revolut-Signature: v1=4fce70bda66b2e713be09fbb7ab1b31b0c8976ea4eeb01b244db7b99aa6482cb,v1=6ffbb59b2300aae63f272406069a9788598b792a944a07aba816edb039989a39

        Raises
        ------
        PyRevolutInvalidPayload
            If the payload signature is invalid or if the payload timestamp is too old.

        Returns
        -------
        dict | ResourceWebhookPayload
            The webhook event payload.
        """
        # Verify the payload signature
        self.verify_payload_signature(
            raw_payload, signing_secret, header_timestamp, header_signature
        )

        # Parse the raw payload
        raw_payload = json.loads(raw_payload)

        # Raw response
        if self.client.return_type == "raw":
            return raw_payload

        # Dict response
        model_response = ResourceWebhookPayload(**raw_payload)
        if self.client.return_type == "dict":
            return model_response.model_dump()

        # Model response
        if self.client.return_type == "model":
            return model_response

    def verify_payload_signature(
        self,
        raw_payload: str,
        signing_secret: str,
        header_timestamp: int,
        header_signature: str,
    ):
        """Verifies the webhook payload signature.

        Parameters
        ----------
        raw_payload : str
            The raw payload string received from the webhook event.
        signing_secret : str
            The signing secret provided by Revolut for the webhook endpoint.
        header_timestamp : int
            The timestamp string received from the webhook event. It
            will be in the header of the request under the key `Revolut-Requested-Timestamp`.
            For example: 1683650202360
        header_signature : str
            Signature of the request payload. It will be in the header of the request
            under the key `Revolut-Signature`.
            Contains the current version of the signature generating algorithm,
            and the hexadecimal-encoded signature itself.
            For example: v1=09a9989dd8d9282c1d34974fc730f5cbfc4f4296941247e90ae5256590a11e8c.

            It can be that there are multiple signatures in the header if multiple signing secrets are active at a given moment.
            If that's the case, they are separated by a comma.
            For example:
            Revolut-Signature: v1=4fce70bda66b2e713be09fbb7ab1b31b0c8976ea4eeb01b244db7b99aa6482cb,v1=6ffbb59b2300aae63f272406069a9788598b792a944a07aba816edb039989a39

        Raises
        ------
        PyRevolutInvalidPayload
            If the payload signature is invalid or if the payload timestamp is too old.

        Returns
        -------
        None
        """
        # Get the revolut signature version from the signature
        revolut_signature_version = header_signature.split("=")[0]

        # Check the timestamp
        current_dt = pendulum.now(tz="UTC")
        timestamp_dt = pendulum.from_timestamp(header_timestamp / 1000, tz="UTC")

        # Check if the timestamp is too old
        if current_dt.diff(timestamp_dt).in_minutes() > 5:
            raise PyRevolutInvalidPayload("The webhook payload timestamp is too old.")

        # Sign the payload
        valid_signature = self.sign_payload(
            raw_payload=raw_payload,
            signing_secret=signing_secret,
            header_timestamp=header_timestamp,
            signature_version=revolut_signature_version,
        )

        # Get the multiple signatures
        multiple_signatures = header_signature.split(",")

        # The computed signature must match exactly the signature (or one of the multiple signatures)
        # sent in that header.
        if valid_signature not in multiple_signatures:
            raise PyRevolutInvalidPayload("The webhook payload signature is invalid.")

    def sign_payload(
        self,
        raw_payload: str,
        signing_secret: str,
        header_timestamp: int,
        signature_version: str = "v1",
    ):
        """Signs the webhook payload.

        Parameters
        ----------
        raw_payload : str
            The raw payload string received from the webhook event.
        signing_secret : str
            The signing secret provided by Revolut for the webhook endpoint.
        header_timestamp : int
            The timestamp string received from the webhook event. It
            will be in the header of the request under the key `Revolut-Requested-Timestamp`.
            For example: 1683650202360
        signature_version : str
            The signature version to use. Default is "v1".

        Returns
        -------
        str
            The signature of the request payload.
        """
        # Create the payload to sign
        payload_to_sign = f"{signature_version}.{header_timestamp}.{raw_payload}"

        # Sign the payload
        return (
            f"{signature_version}="
            + hmac.new(
                bytes(signing_secret, "utf-8"),
                msg=bytes(payload_to_sign, "utf-8"),
                digestmod=hashlib.sha256,
            ).hexdigest()
        )



from uuid import UUID

from pyrevolut.api.common import BaseEndpointSync

from pyrevolut.api.accounts.get import (
    RetrieveAllAccounts,
    RetrieveAnAccount,
    RetrieveFullBankDetails,
)


class EndpointAccountsSync(BaseEndpointSync):
    """The Accounts API
    Get the balances, full banking details, and other details of your business accounts.
    """

    def get_all_accounts(
        self,
        **kwargs,
    ) -> list[dict] | list[RetrieveAllAccounts.Response]:
        """
        Get a list of all your accounts.

        Parameters
        ----------
        None

        Returns
        -------
        list[dict] | list[RetrieveAllAccounts.Response]
            The list of all your accounts
        """
        endpoint = RetrieveAllAccounts
        path = endpoint.ROUTE
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_account(
        self,
        account_id: UUID,
        **kwargs,
    ) -> dict | RetrieveAnAccount.Response:
        """
        Get the information about one of your accounts. Specify the account by its ID.

        Parameters
        ----------
        account_id : UUID
            The account ID.

        Returns
        -------
        dict | RetrieveAnAccount.Response
            The information about the account
        """
        endpoint = RetrieveAnAccount
        path = endpoint.ROUTE.format(account_id=account_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    def get_full_bank_details(
        self,
        account_id: UUID,
        **kwargs,
    ) -> dict | RetrieveFullBankDetails.Response:
        """
        Get all the bank details of one of your accounts. Specify the account by its ID.

        Parameters
        ----------
        account_id : UUID
            The account ID.

        Returns
        -------
        dict | RetrieveFullBankDetails.Response
            The bank details of the account
        """
        endpoint = RetrieveFullBankDetails
        path = endpoint.ROUTE.format(account_id=account_id)
        params = endpoint.Params()

        return self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )



from pydantic import BaseModel

from pyrevolut.api.payout_links.resources import ResourcePayoutLink


class RetrievePayoutLink:
    """
    Get the information about a specific link by its ID.

    Note
    ----
    This feature is available in the UK and the EEA.
    """

    ROUTE = "/1.0/payout-links/{payout_link_id}"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        pass

    class Response(ResourcePayoutLink):
        """
        The response model for the request.
        """

        pass



from uuid import UUID
from datetime import datetime

from pydantic_extra_types.pendulum_dt import Duration

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumWebhookEvent

from pyrevolut.api.webhooks.get import (
    RetrieveListOfWebhooks,
    RetrieveWebhook,
    RetrieveListOfFailedWebhooks,
)
from pyrevolut.api.webhooks.post import CreateWebhook, RotateWebhookSecret
from pyrevolut.api.webhooks.patch import UpdateWebhook
from pyrevolut.api.webhooks.delete import DeleteWebhook

from .base import BaseEndpointWebhooks


class EndpointWebhooksAsync(BaseEndpointWebhooks):
    """The async Webhooks API

    A webhook (also called a web callback) allows your system to receive
    updates about your account to an HTTPS endpoint that you provide.
    When a supported event occurs, a notification is posted via HTTP POST method
    to the specified endpoint.

    If the receiver returns an HTTP error response, Revolut will retry the webhook
    event three more times, each with a 10-minute interval.

    The following events are supported:

    TransactionCreated
    TransactionStateChanged
    PayoutLinkCreated
    PayoutLinkStateChanged
    """

    async def get_all_webhooks(
        self,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfWebhooks.Response]:
        """
        Get the list of all your existing webhooks and their details.

        Parameters
        ----------
        None

        Returns
        -------
        list[dict] | list[RetrieveListOfWebhooks.Response]
            The list of all your existing webhooks and their details.
        """
        endpoint = RetrieveListOfWebhooks
        path = endpoint.ROUTE
        params = endpoint.Params()

        return await self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    async def get_webhook(
        self,
        webhook_id: UUID,
        **kwargs,
    ) -> dict | RetrieveWebhook.Response:
        """
        Get the information about a specific webhook by ID.

        Parameters
        ----------
        webhook_id : UUID
            The ID of the webhook.

        Returns
        -------
        dict | RetrieveWebhook.Response
            The information about the webhook.
        """
        endpoint = RetrieveWebhook
        path = endpoint.ROUTE.format(webhook_id=webhook_id)
        params = endpoint.Params()

        return await self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    async def get_failed_webhook_events(
        self,
        webhook_id: UUID,
        limit: int | None = None,
        created_before: datetime | DateTime | str | int | float | None = None,
        **kwargs,
    ) -> list[dict] | list[RetrieveListOfFailedWebhooks.Response]:
        """
        Get the list of all your failed webhook events, or use the query
        parameters to filter the results.

        The events are sorted by the created_at date in reverse chronological order.

        The returned failed events are paginated. The maximum number of events returned
        per page is specified by the limit parameter.
        To get to the next page, make a new request and use the created_at date of the
        last event returned in the previous response.

        Parameters
        ----------
        webhook_id : UUID
            The ID of the webhook.
        limit : int, optional
            The maximum number of events returned per page.
            To get to the next page, make a new request and use the created_at date of
            the last event returned in the previous response as value for created_before.
            If not specified, the default value is 100.
        created_before : datetime | DateTime | str | int | float, optional
            Retrieves events with created_at < created_before.
            Cannot be older than the current date minus 21 days.
            The default value is the current date and time at which you are calling the endpoint.
            Provided in ISO 8601 format.

        Returns
        -------
        list[dict] | list[RetrieveListOfFailedWebhooks.Response]
            The list of all your failed webhook events.
        """
        endpoint = RetrieveListOfFailedWebhooks
        path = endpoint.ROUTE.format(webhook_id=webhook_id)
        params = endpoint.Params(
            limit=limit,
            created_before=created_before,
        )

        return await self.client.get(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )

    async def create_webhook(
        self,
        url: str,
        events: list[EnumWebhookEvent] | None = None,
        **kwargs,
    ) -> dict | CreateWebhook.Response:
        """
        Create a new webhook to receive event notifications to the specified URL.
        Provide a list of event types that you want to subscribe to and a URL for the webhook.
        Only HTTPS URLs are supported.

        Parameters
        ----------
        url : str
            A valid webhook URL to which to send event notifications.
            The supported protocol is https.
        events : list[EnumWebhookEvent], optional
            A list of event types to subscribe to.
            If you don't provide it, you're automatically subscribed to the default event types:
            - TransactionCreated
            - TransactionStateChanged

        Returns
        -------
        dict | CreateWebhook.Response
            The response model for the request.
        """
        endpoint = CreateWebhook
        path = endpoint.ROUTE
        body = endpoint.Body(
            url=url,
            events=events,
        )

        return await self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    async def rotate_webhook_secret(
        self,
        webhook_id: UUID,
        expiration_period: Duration | None = None,
        **kwargs,
    ) -> dict | RotateWebhookSecret.Response:
        """
        Rotate a signing secret for a specific webhook.

        Parameters
        ----------
        webhook_id : UUID
            The ID of the webhook.
        expiration_period : Duration, optional
            The expiration period for the signing secret in ISO 8601 format.
            If set, when you rotate the secret, it continues to be valid until the
            expiration period has passed.
            Otherwise, on rotation, the secret is invalidated immediately.
            The maximum value is 7 days.

        Returns
        -------
        dict | RotateWebhookSecret.Response
            The response model for the request.
        """
        endpoint = RotateWebhookSecret
        path = endpoint.ROUTE.format(webhook_id=webhook_id)
        body = endpoint.Body(
            expiration_period=expiration_period,
        )

        return await self.client.post(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    async def update_webhook(
        self,
        webhook_id: UUID,
        url: str | None = None,
        events: list[EnumWebhookEvent] | None = None,
        **kwargs,
    ) -> dict | UpdateWebhook.Response:
        """
        Update an existing webhook. Change the URL to which event notifications are
        sent or the list of event types to be notified about.

        You must specify at least one of these two.
        The fields that you don't specify are not updated.

        Parameters
        ----------
        webhook_id : UUID
            The ID of the webhook.
        url : str, optional
            A valid webhook URL to which to send event notifications.
            The supported protocol is https.
        events : list[EnumWebhookEvent], optional
            A list of event types to subscribe to.

        Returns
        -------
        dict | UpdateWebhook.Response
            The response model for the request.
        """
        endpoint = UpdateWebhook
        path = endpoint.ROUTE.format(webhook_id=webhook_id)
        body = endpoint.Body(
            url=url,
            events=events,
        )

        return await self.client.patch(
            path=path,
            response_model=endpoint.Response,
            body=body,
            **kwargs,
        )

    async def delete_webhook(
        self,
        webhook_id: UUID,
        **kwargs,
    ) -> dict | DeleteWebhook.Response:
        """
        Delete a specific webhook.

        A successful response does not get any content in return.

        Parameters
        ----------
        webhook_id : UUID
            The ID of the webhook.

        Returns
        -------
        dict | DeleteWebhook.Response
            An empty dictionary.
        """
        endpoint = DeleteWebhook
        path = endpoint.ROUTE.format(webhook_id=webhook_id)
        params = endpoint.Params()

        return await self.client.delete(
            path=path,
            response_model=endpoint.Response,
            params=params,
            **kwargs,
        )



from typing import Annotated

from pydantic import BaseModel, Field, ConfigDict

from pyrevolut.utils import DateTime
from pyrevolut.api.common import ModelBaseAmount


class ResourceForeignExchange(BaseModel):
    """
    Foreign Exchange resource model.
    """

    model_config = ConfigDict(
        populate_by_name=True,
        from_attributes=True,
    )

    class ModelFrom(ModelBaseAmount):
        """
        The money to sell.
        """

        pass

    class ModelTo(ModelBaseAmount):
        """
        The money to receive.
        """

        pass

    class ModelFee(ModelBaseAmount):
        """
        The expected fee for the transaction.
        """

        pass

    from_: Annotated[ModelFrom, Field(alias="from", description="The money to sell.")]
    to: Annotated[ModelTo, Field(description="The money to receive.")]
    rate: Annotated[float, Field(description="The proposed exchange rate.")]
    fee: Annotated[ModelFee, Field(description="The expected fee for the transaction.")]
    rate_date: Annotated[
        DateTime,
        Field(description="The date of the proposed exchange rate in ISO 8601 format."),
    ]



from typing import Annotated, Literal

from pydantic import BaseModel, Field, model_validator

from pyrevolut.api.common import ModelBaseAmount, EnumMerchantCategory
from pyrevolut.api.cards.resources import ResourceCard


class UpdateCardDetails:
    """
    Update details of a specific card, based on its ID.
    Updating a spending limit does not reset the spending counter.
    """

    ROUTE = "/1.0/cards/{card_id}"

    class Body(BaseModel):
        """
        The body of the request.
        """

        class ModelSpendingLimits(BaseModel):
            """All spending limits set for the card."""

            class ModelSingle(ModelBaseAmount):
                """The limit for a single transaction."""

                pass

            class ModelDay(ModelBaseAmount):
                """The daily limit for transactions."""

                pass

            class ModelWeek(ModelBaseAmount):
                """The weekly limit for transactions."""

                pass

            class ModelMonth(ModelBaseAmount):
                """The monthly limit for transactions."""

                pass

            class ModelQuarter(ModelBaseAmount):
                """The quarterly limit for transactions."""

                pass

            class ModelYear(ModelBaseAmount):
                """The yearly limit for transactions."""

                pass

            class ModelAllTime(ModelBaseAmount):
                """The all-time limit for transactions."""

                pass

            single: Annotated[
                ModelSingle | Literal["null"] | None,
                Field(description="The limit for a single transaction."),
            ]
            day: Annotated[
                ModelDay | Literal["null"] | None,
                Field(description="The daily limit for transactions."),
            ]
            week: Annotated[
                ModelWeek | Literal["null"] | None,
                Field(description="The weekly limit for transactions."),
            ]
            month: Annotated[
                ModelMonth | Literal["null"] | None,
                Field(description="The monthly limit for transactions."),
            ]
            quarter: Annotated[
                ModelQuarter | Literal["null"] | None,
                Field(description="The quarterly limit for transactions."),
            ]
            year: Annotated[
                ModelYear | Literal["null"] | None,
                Field(description="The yearly limit for transactions."),
            ]
            all_time: Annotated[
                ModelAllTime | Literal["null"] | None,
                Field(description="The all-time limit for transactions."),
            ]

        label: Annotated[
            str | None,
            Field(description="The label of the card.", max_length=30),
        ] = None
        categories: Annotated[
            EnumMerchantCategory | Literal["null"] | None,
            Field(
                description="""
                List of merchant categories that will be available for card spending. 
                Use null to erase the value and reset to empty (all categories will be allowed).
                """
            ),
        ] = None
        spending_limits: Annotated[
            ModelSpendingLimits | Literal["null"] | None,
            Field(
                description="""
                All spending limits set for the card.

                You can have at most 1 periodic (day/week/month/quarter/all-time) and 1 non-periodic (single transaction) 
                limit at a time. If you try to specify 2 periodic limits at a time, it will result in an error.

                Spending limit currency must match the default business currency. 
                The default currency was assigned to your business during onboarding.

                Use null as the value for a specific limit to erase that limit. 
                Use null as the value for the spending_limits object to erase all limits.
                """
            ),
        ] = None

        @model_validator(mode="after")
        def check_inputs(self) -> "UpdateCardDetails.Body":
            """Check the inputs."""
            if self.spending_limits is not None and self.spending_limits != "null":
                if (
                    sum(
                        [
                            self.spending_limits.day is not None
                            and self.spending_limits.day != "null",
                            self.spending_limits.week is not None
                            and self.spending_limits.week != "null",
                            self.spending_limits.month is not None
                            and self.spending_limits.month != "null",
                            self.spending_limits.quarter is not None
                            and self.spending_limits.quarter != "null",
                            self.spending_limits.year is not None
                            and self.spending_limits.year != "null",
                            self.spending_limits.all_time is not None
                            and self.spending_limits.all_time != "null",
                        ]
                    )
                    > 1
                ):
                    raise ValueError(
                        "You can have at most 1 periodic (day/week/month/quarter/all-time) limit at a time."
                    )
            return self

    class Response(ResourceCard):
        """
        Response model for the endpoint.
        """

        pass



"""This module holds all GET payout links endpoints."""

# flake8: noqa: F401
from .retrieve_list_of_payout_links import RetrieveListOfPayoutLinks
from .retrieve_payout_link import RetrievePayoutLink



"""This module holds all GET accounts endpoints."""

# flake8: noqa: F401
from .retrieve_all_accounts import RetrieveAllAccounts
from .retrieve_an_account import RetrieveAnAccount
from .retrieve_full_bank_details import RetrieveFullBankDetails



from .common import PyRevolutBaseException


class PyRevolutForbidden(PyRevolutBaseException):
    """Forbidden -- Access to the requested resource or action is forbidden."""

    def __init__(
        self,
        msg="Access to the requested resource or action is forbidden.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic_extra_types.currency_code import Currency

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumAccountState


class ResourceAccount(BaseModel):
    """
    Account resource model.
    """

    id: Annotated[
        UUID,
        Field(description="The account ID."),
    ]
    name: Annotated[
        str | None,
        Field(description="The account name."),
    ] = None
    balance: Annotated[
        float,
        Field(description="The current balance on the account."),
    ]
    currency: Annotated[
        Currency,
        Field(description="ISO 4217 currency code in upper case."),
    ]
    state: Annotated[
        EnumAccountState,
        Field(description="Indicates the state of the account."),
    ]
    public: Annotated[
        bool,
        Field(
            description="Indicates whether the account is visible to other businesses on Revolut."
        ),
    ]
    created_at: Annotated[
        DateTime,
        Field(
            description="The date and time the account was created in ISO 8601 format."
        ),
    ]
    updated_at: Annotated[
        DateTime,
        Field(
            description="The date and time the account was last updated in ISO 8601 format."
        ),
    ]



from pydantic import BaseModel


class DeletePaymentDraft:
    """
    Delete a payment draft with the given ID.
    You can delete a payment draft only if it isn't processed.
    """

    ROUTE = "/1.0/payment-drafts/{payment_draft_id}"

    class Params(BaseModel):
        """
        The parameters for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        The response model for the endpoint.
        """

        pass



from pydantic import BaseModel

from pyrevolut.api.counterparties.resources import ResourceCounterparty


class RetrieveCounterparty:
    """Get the information about a specific counterparty by ID."""

    ROUTE = "/1.0/counterparty/{counterparty_id}"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        pass

    class Response(ResourceCounterparty):
        """
        Response model for the endpoint.
        """

        pass



from enum import StrEnum


class EnumCardState(StrEnum):
    """Card state enumeration"""

    CREATED = "created"
    PENDING = "pending"
    ACTIVE = "active"
    FROZEN = "frozen"
    LOCKED = "locked"



from typing import Annotated

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime
from pyrevolut.api.cards.resources import ResourceCard


class RetrieveListOfCards:
    """
    Get the list of all cards in your organisation.
    The results are paginated and sorted by the created_at date in reverse chronological order.
    """

    ROUTE = "/1.0/cards"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        created_before: Annotated[
            DateTime | None,
            Field(
                description="""
                Retrieves cards with created_at < created_before. 
                The default value is the current date and time at which you are calling the endpoint.
                Provided in ISO 8601 format.
                """
            ),
        ] = None
        limit: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of cards returned per page.
                To get to the next page, make a new request and use the 
                created_at date of the last card returned in the previous 
                response as the value for created_before.              
                
                If not provided, the default value is 100.  
                """,
                ge=1,
                le=100,
            ),
        ] = None

    class Response(ResourceCard):
        """
        Response model for the endpoint.
        """

        pass



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime


class RetrieveTeamRoles:
    """
    Get the list of roles for your business.

    The results are paginated and sorted by the created_at date in reverse chronological order.

    This feature is available in the UK, US and the EEA.

    This feature is not available in Sandbox.
    """

    ROUTE = "/1.0/roles"

    class Params(BaseModel):
        """
        The query parameters of the request.
        """

        created_before: Annotated[
            DateTime | None,
            Field(
                description="""
                Retrieves roles with created_at < created_before. 
                The default value is the current date and time at which you are calling the endpoint.
                Provided in ISO 8601 format.
                """
            ),
        ] = None
        limit: Annotated[
            int | None,
            Field(
                description="""
                The maximum number of roles returned per page.
                To get to the next page, make a new request and use the 
                created_at date of the last role returned in the previous 
                response as the value for created_before.              
                
                If not provided, the default value is 100.  
                """,
                ge=1,
                le=100,
            ),
        ] = None

    class Response(BaseModel):
        """
        The response model.
        """

        id: Annotated[
            UUID | str,
            Field(
                description="The ID of the role. This can be a UUID or other default role such as OWNER."
            ),
        ]
        name: Annotated[str, Field(description="The name of the role.")]
        created_at: Annotated[
            DateTime,
            Field(
                description="The date and time the role was created in ISO 8601 format."
            ),
        ]
        updated_at: Annotated[
            DateTime,
            Field(
                description="The date and time the role was last updated in ISO 8601 format."
            ),
        ]



from typing import Type, Any
import datetime

import pendulum
from pendulum import DateTime as _DateTime  # type: ignore
from pendulum.exceptions import PendulumException

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class DateTime(_DateTime):
    """
    A `pendulum.DateTime` object. At runtime, this type decomposes into pendulum.DateTime automatically.
    This type exists because Pydantic throws a fit on unknown types.
    """

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Return a Pydantic CoreSchema with the DateTime validation

        Parameters
        ----------
        source: Type[Any]
            The source type to be converted.
        handler: GetCoreSchemaHandler
            The handler to get the CoreSchema.

        Returns
        -------
        core_schema.CoreSchema
            A Pydantic CoreSchema with the DateTime validation.
        """
        return core_schema.no_info_wrap_validator_function(
            cls._validate, core_schema.datetime_schema()
        )

    @classmethod
    def _validate(
        cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler
    ) -> Any:
        """
        Validate the datetime object and return it.

        Parameters
        ----------
        value: Any
            The value to validate.
        handler: core_schema.ValidatorFunctionWrapHandler
            The handler to get the CoreSchema.

        Returns
        -------
        Any
            The validated value or raises a PydanticCustomError.
        """
        try:
            data = to_datetime(value)
        except Exception as exc:
            raise PydanticCustomError("value_error", str(exc)) from exc  # type: ignore
        return handler(data)


def to_datetime(
    dt: datetime.datetime | _DateTime | str | int | float, timestamp_unit: str = "ms"
):
    """
    Converts a datetime object or string to a pendulum DateTime object.

    Parameters
    ----------
    dt : datetime | DateTime | str | int | float
        The object to be converted.
    timestamp_unit : str, optional
        The unit of the timestamp. The default is 'ms'.

    Returns
    -------
    DateTime
        The pendulum DateTime object.

    """
    if isinstance(dt, datetime.datetime):
        new_dt: DateTime = pendulum.instance(dt, tz="UTC")  # type: ignore
    elif isinstance(dt, str):
        new_dt: DateTime = string_to_datetime(dt)
    elif isinstance(dt, int) or isinstance(dt, float):
        new_dt: DateTime = timestamp_to_datetime(dt, unit=timestamp_unit)
    new_dt: DateTime = pendulum.instance(new_dt, tz="UTC")  # type: ignore
    return new_dt


def timestamp_to_datetime(
    timestamp: float | int,
    unit: str = "ms",
):
    """
    Converts a timestamp to a pendulum DateTime object.

    Parameters
    ----------
    timestamp : float | int
        The timestamp to be converted.
    unit : str, optional
        The unit of the timestamp. The default is 'ms'.

    Returns
    -------
    DateTime
        The pendulum DateTime object.

    Raises
    ------
    ValueError
        If the unit is not supported.

    """
    if unit == "ms":
        dt: DateTime = pendulum.from_timestamp(timestamp / 1000, tz="UTC")  # type: ignore
    elif unit == "s":
        dt: DateTime = pendulum.from_timestamp(timestamp, tz="UTC")  # type: ignore
    else:
        raise ValueError(f"Unsupported unit: {unit}")
    return dt


def string_to_datetime(string: str):
    """
    Converts a string to a pendulum DateTime object.

    Parameters
    ----------
    string : str
        The string to be converted.

    Returns
    -------
    DateTime
        The pendulum DateTime object.

    Raises
    ------
    ValueError
        If the string cannot be converted to a pendulum DateTime object.

    """
    try:
        dt: DateTime = pendulum.parser.parse(string).in_timezone("UTC")  # type: ignore
        return dt
    except PendulumException as exc:
        raise ValueError(f"Could not parse string: {string}") from exc



from enum import StrEnum


class EnumPayoutLinkPaymentMethod(StrEnum):
    """Payout link payment method.

    The list of payout methods that the recipient can use to claim the payout, where:

        revolut:
            Revolut peer-to-peer (P2P) transfer

        bank_account:
            External bank transfer

        card:
            Card transfer
    """

    REVOLUT = "revolut"
    BANK_ACCOUNT = "bank_account"
    CARD = "card"



from typing import Annotated

from pydantic import BaseModel, Field
from pydantic_extra_types.pendulum_dt import Duration

from pyrevolut.api.webhooks.resources import ResourceWebhook


class RotateWebhookSecret:
    """
    Rotate a signing secret for a specific webhook.
    """

    ROUTE = "/2.0/webhooks/{webhook_id}/rotate-signing-secret"

    class Body(BaseModel):
        """
        The body of the request.
        """

        expiration_period: Annotated[
            Duration | None,
            Field(
                description="""
                The expiration period for the signing secret in ISO 8601 format. 
                If set, when you rotate the secret, it continues to be valid until the 
                expiration period has passed. Otherwise, on rotation, the secret is 
                invalidated immediately. 
                The maximum value is 7 days.
                """,
            ),
        ] = None

    class Response(ResourceWebhook):
        """
        The response model for the request.
        """

        signing_secret: Annotated[
            str,
            Field(
                description="""
                The signing secret for the webhook.
                """,
            ),
        ]



from pydantic import BaseModel


class FreezeCard:
    """
    Freeze a card to make it temporarily unavailable for spending.
    You can only freeze a card that is in the state active.

    A successful freeze changes the card's state to frozen,
    and no content is returned in the response.
    """

    ROUTE = "/1.0/cards/{card_id}/freeze"

    class Body(BaseModel):
        """
        Request body for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        pass



from enum import StrEnum


class EnumPayoutLinkCancellationReason(StrEnum):
    """Payout link cancellation reason.

    The reason why the payout link was cancelled. Possible reasons are:

    too_many_name_check_attempts:
        The name check failed too many times.
    """

    TOO_MANY_NAME_CHECK_ATTEMPTS = "too_many_name_check_attempts"



from typing import Annotated

from pydantic import BaseModel, Field, ConfigDict
from pydantic_extra_types.currency_code import Currency

from pyrevolut.api.foreign_exchange.resources import ResourceForeignExchange


class GetExchangeRate:
    """
    Get the sell exchange rate between two currencies.
    """

    ROUTE = "/1.0/rate"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        model_config = ConfigDict(
            populate_by_name=True,
            from_attributes=True,
        )

        from_: Annotated[
            Currency,
            Field(
                alias="from",
                description="The currency that you exchange from in ISO 4217 format.",
            ),
        ]
        to: Annotated[
            Currency,
            Field(description="The currency that you exchange to in ISO 4217 format."),
        ]
        amount: Annotated[
            float | None,
            Field(
                description="The amount of the currency to exchange from. The default value is 1.00."
            ),
        ] = None

    class Response(ResourceForeignExchange):
        """
        Response model for the endpoint.
        """

        pass



from enum import StrEnum


class EnumTransactionState(StrEnum):
    """
    Indicates the transaction state. Possible values:

    created:
        The transaction has been created and is either processed asynchronously
        or scheduled for a later time.
    pending:
        The transaction is pending until it's being processed.
        If the transfer is made between Revolut accounts,
        this state is skipped and the transaction is executed instantly.
    completed:
        The transaction was successful.
    declined:
        The transaction was unsuccessful. This can happen for a variety of reasons,
        for example, insufficient account balance, wrong receiver information, etc.
    failed:
        The transaction was unsuccessful. This can happen for a variety of reasons,
        for example, invalid API calls, blocked payments, etc.
    reverted:
        The transaction was reverted. This can happen for a variety of reasons,
        for example, the receiver being inaccessible.
    cancelled:
        The transaction was cancelled.
    """

    CREATED = "created"
    PENDING = "pending"
    COMPLETED = "completed"
    DECLINED = "declined"
    FAILED = "failed"
    REVERTED = "reverted"
    CANCELLED = "cancelled"



from .common import PyRevolutBaseException


class PyRevolutUnauthorized(PyRevolutBaseException):
    """Unauthorized -- Your access token is wrong."""

    def __init__(
        self,
        msg="Your access token is wrong. Please check the access token and try again.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



"""Retrieve information on existing team members of your organisation and invite new members.

This feature is available in the UK, US and the EEA.

This feature is not available in Sandbox.

Reference link: https://developer.revolut.com/docs/business/team-members
"""

# flake8: noqa: F401
from .endpoint import EndpointTeamMembersSync, EndpointTeamMembersAsync



"""This module holds all GET counterparties endpoints."""

# flake8: noqa: F401
from .retrieve_counterparty import RetrieveCounterparty
from .retrieve_list_of_counterparties import RetrieveListOfCounterparties



from pydantic import BaseModel


class CancelPayoutLink:
    """
    Cancel a payout link.
    You can only cancel a link that hasn't been claimed yet.
    A successful request does not get any content in response.

    Note
    ----
    This feature is available in the UK and the EEA.
    """

    ROUTE = "/1.0/payout-links/{payout_link_id}/cancel"

    class Body(BaseModel):
        """
        The request body model for the request.
        """

        pass

    class Response(BaseModel):
        """
        The response model for the request.
        """

        pass



from typing import Annotated

from pydantic import BaseModel, Field, HttpUrl

from pyrevolut.api.common import EnumWebhookEvent
from pyrevolut.api.webhooks.resources import ResourceWebhook


class UpdateWebhook:
    """
    Update an existing webhook. Change the URL to which event notifications are
    sent or the list of event types to be notified about.

    You must specify at least one of these two.
    The fields that you don't specify are not updated.
    """

    ROUTE = "/2.0/webhooks/{webhook_id}"

    class Body(BaseModel):
        """
        The body of the request.
        """

        url: Annotated[
            HttpUrl | None,
            Field(
                description="""
                A valid webhook URL to which to send event notifications. 
                The supported protocol is https.
                """,
            ),
        ] = None
        events: Annotated[
            list[EnumWebhookEvent] | None,
            Field(
                description="""
                A list of event types to subscribe to. 
                """,
            ),
        ] = None

    class Response(ResourceWebhook):
        """
        The response model for the request.
        """

        pass



from typing import Annotated

from pydantic import BaseModel, Field

from pyrevolut.api.webhooks.resources import ResourceWebhook


class RetrieveWebhook:
    """
    Get the information about a specific webhook by ID.
    """

    ROUTE = "/2.0/webhooks/{webhook_id}"

    class Params(BaseModel):
        """
        The query parameters for the request.
        """

        pass

    class Response(ResourceWebhook):
        """
        The response model for the request.
        """

        signing_secret: Annotated[
            str,
            Field(
                description="""
                The signing secret for the webhook.
                """,
            ),
        ]



from typing import Annotated

from httpx import Client, AsyncClient
from pydantic import BaseModel, Field, SecretStr


class ModelGetAuthTokensResponse(BaseModel):
    """The model that represents the response from the get auth tokens endpoint."""

    access_token: Annotated[SecretStr, Field(description="The access token")]
    refresh_token: Annotated[SecretStr, Field(description="The refresh token")]
    token_type: Annotated[str, Field(description="The token type")]
    expires_in: Annotated[int, Field(description="The expiration time in seconds")]


def get_auth_tokens(
    client: Client,
    auth_code: str,
    client_assert_jwt: str,
    sandbox: bool = True,
):
    """
    Method to get the access and refresh token from the Revolut API.

    Parameters
    ----------
    client : Client
        The HTTPX client.
    auth_code : str
        The authorization code.
    client_assert_jwt : str
        The client assertion JWT.
    sandbox : bool, optional
        Whether to use the sandbox environment.
        Default is True.

    Returns
    -------
    dict
        The access and refresh token response.
    """
    response = client.post(
        **prep_get_auth_tokens(
            auth_code=auth_code,
            client_assert_jwt=client_assert_jwt,
            sandbox=sandbox,
        ),
    )
    return ModelGetAuthTokensResponse(**response.json())


async def aget_auth_tokens(
    client: AsyncClient,
    auth_code: str,
    client_assert_jwt: str,
    sandbox: bool = True,
):
    """
    Method to get the access and refresh token from the Revolut API.

    Parameters
    ----------
    client : Client
        The async HTTPX client.
    auth_code : str
        The authorization code.
    client_assert_jwt : str
        The client assertion JWT.
    sandbox : bool, optional
        Whether to use the sandbox environment.
        Default is True.

    Returns
    -------
    dict
        The access and refresh token response.
    """
    response = await client.post(
        **prep_get_auth_tokens(
            auth_code=auth_code,
            client_assert_jwt=client_assert_jwt,
            sandbox=sandbox,
        ),
    )
    return ModelGetAuthTokensResponse(**response.json())


def prep_get_auth_tokens(
    auth_code: str,
    client_assert_jwt: str,
    sandbox: bool = True,
):
    """
    Method to prepare the arguments for getting the auth token functions.

    Parameters
    ----------
    auth_code : str
        The authorization code.
    client_assert_jwt : str
        The client assertion JWT.
    sandbox : bool, optional
        Whether to use the sandbox environment.
        Default is True.

    Returns
    -------
    dict
        The arguments to be passed to the HTTPX client POST method.
    """
    if sandbox:
        url = "https://sandbox-b2b.revolut.com/api/1.0/auth/token"
    else:
        url = "https://b2b.revolut.com/api/1.0/auth/token"
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    data = {
        "grant_type": "authorization_code",
        "code": auth_code,
        "client_id": "client_id",
        "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
        "client_assertion": client_assert_jwt,
    }
    return {
        "url": url,
        "headers": headers,
        "data": data,
    }



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, model_validator
from pydantic_extra_types.currency_code import Currency

from pyrevolut.utils import DateTime
from pyrevolut.api.common import EnumTransactionState


class SimulateAccountTopup:
    """
    Simulate a top-up of your account in the Sandbox environment.

    This is useful during testing, when you run out of money in your test account
    and need to add more.
    """

    ROUTE = "/1.0/sandbox/topup"

    class Body(BaseModel):
        """
        The body of the request.
        """

        account_id: Annotated[
            UUID,
            Field(description="The ID of the account that you want to top up."),
        ]
        amount: Annotated[
            float,
            Field(
                description="The amount with which you want to top up the account. Must be <= 10000",
                gt=0,
                le=10000,
            ),
        ]
        currency: Annotated[
            Currency,
            Field(
                description="The currency of the top-up amount. Must be a valid ISO 4217 currency code.",
            ),
        ]
        reference: Annotated[
            str | None,
            Field(
                description="""
                A short description for your top up.
                Default value: 'Test Top-up' if not provided.
                """,
            ),
        ] = None
        state: Annotated[
            EnumTransactionState | None,
            Field(
                description="""
                The state to which you want to set the top-up transaction. 
                
                If not provided, the default value is 'completed'.
                
                Possible values:

                    pending:
                        The transaction is pending until it's being processed.
                        If the transfer is made between Revolut accounts,
                        this state is skipped and the transaction is executed instantly.
                    completed:
                        The transaction was successful.
                    failed:
                        The transaction was unsuccessful. This can happen for a variety of reasons,
                        for example, invalid API calls, blocked payments, etc.
                    reverted:
                        The transaction was reverted. This can happen for a variety of reasons,
                        for example, the receiver being inaccessible.
                """
            ),
        ] = None

        @model_validator(mode="after")
        def check_inputs(self) -> "SimulateAccountTopup.Body":
            """
            Check that the input is correct.
            """

            if self.state is not None:
                assert self.state in [
                    EnumTransactionState.PENDING,
                    EnumTransactionState.COMPLETED,
                    EnumTransactionState.FAILED,
                    EnumTransactionState.REVERTED,
                ], (
                    f"Invalid state: {self.state}. "
                    f"Possible values: {EnumTransactionState.PENDING}, "
                    f"{EnumTransactionState.COMPLETED}, {EnumTransactionState.FAILED}, "
                    f"{EnumTransactionState.REVERTED}."
                )
            return self

    class Response(BaseModel):
        """
        The response model.
        """

        id: Annotated[
            UUID, Field(description="The ID of the account that was topped up.")
        ]
        state: Annotated[
            EnumTransactionState,
            Field(
                description="""
                The state of the top-up transaction. Possible values:

                pending:
                    The transaction is pending until it's being processed.
                    If the transfer is made between Revolut accounts,
                    this state is skipped and the transaction is executed instantly.
                completed:
                    The transaction was successful.
                failed:
                    The transaction was unsuccessful. This can happen for a variety of reasons,
                    for example, invalid API calls, blocked payments, etc.
                reverted:
                    The transaction was reverted. This can happen for a variety of reasons,
                    for example, the receiver being inaccessible.
                """
            ),
        ]
        created_at: Annotated[
            DateTime,
            Field(
                description="The date and time the transaction was created in ISO 8601 format."
            ),
        ]
        completed_at: Annotated[
            DateTime | None,
            Field(
                description="The date and time the transaction was completed in ISO 8601 format."
            ),
        ] = None



import os
from typing import Annotated

import typer

from pydantic import BaseModel
from pyrevolut.utils.auth import EnumAuthScope, auth_manual_flow

app = typer.Typer()


class AuthManualParams(BaseModel):
    """Pydantic model for the auth_manual CLI command."""

    credentials_json: str
    sandbox: bool = True
    scopes: list[EnumAuthScope] | None = None


@app.callback()
def callback():
    """
    pyrevolut CLI tool to interact with the Revolut Business API. Primarily
    used for authentication and authorization.
    """


@app.command(name="auth-manual")
def auth_manual(
    credentials_json: str = "credentials/creds.json",
    sandbox: bool = True,
    scopes: Annotated[list[str], typer.Option()] = None,
):
    """
    Method to run the manual authorization flow to get the client assertion JWT and the access and refresh tokens.

    Parameters
    ----------
    credentials_json : str, optional
        The location to save the credentials JSON file.
        Will overwrite the file if it already exists.
        Default is "credentials/creds.json".
    sandbox : bool, optional
        Whether to use the sandbox environment.
        Default is True.
    scopes : list[str] | None, optional
        The list of scopes to request. If not provided, the default scopes will be used.

        Access tokens can be issued with four security scopes and require a JWT (JSON Web Token)
        signature to be obtained:

            READ: Permissions for GET operations.

            WRITE: Permissions to update counterparties, webhooks, and issue payment drafts.

            PAY: Permissions to initiate or cancel transactions and currency exchanges.

            READ_SENSITIVE_CARD_DATA: Permissions to retrieve sensitive card details.

        Caution
        -------
        If you enable the READ_SENSITIVE_CARD_DATA scope for your access token, you must
        set up IP whitelisting.
        Failing to do so will prevent you from accessing any Business API endpoint.

        Default is None.

    Returns
    -------
    None
    """
    params = AuthManualParams(
        credentials_json=credentials_json,
        sandbox=sandbox,
        scopes=scopes,
    )

    # Create the directories if they do not exist
    os.makedirs(os.path.dirname(params.credentials_json), exist_ok=True)

    # Run the manual authorization flow
    auth_manual_flow(
        credentials_json=params.credentials_json,
        sandbox=params.sandbox,
        scopes=params.scopes,
    )



from enum import StrEnum


class EnumTransferReasonCode(StrEnum):
    """Transfer reason code."""

    ADVERTISING = "advertising"
    ADVISOR_FEES = "advisor_fees"
    BUSINESS_INSURANCE = "business_insurance"
    CONSTRUCTION = "construction"
    DELIVERY = "delivery"
    EDUCATION = "education"
    FAMILY = "family"
    FUND_INVESTMENT = "fund_investment"
    GOODS = "goods"
    HOMESEND = "homesend"
    HOTEL = "hotel"
    EXPORTS = "exports"
    INSURANCE_CLAIMS = "insurance_claims"
    INSURANCE_PREMIUM = "insurance_premium"
    LOAN_REPAYMENT = "loan_repayment"
    MEDICAL = "medical"
    OFFICE = "office"
    OTHER_FEES = "other_fees"
    PERSONAL_TRANSFER = "personal_transfer"
    PROPERTY_PURCHASE = "property_purchase"
    PROPERTY_RENTAL = "property_rental"
    ROYALTIES = "royalties"
    SERVICES = "services"
    SHARE_INVESTMENT = "share_investment"
    TAX = "tax"
    TRANSFER = "transfer"
    TRANSPORTATION = "transportation"
    TRAVEL = "travel"
    UTILITIES = "utilities"
    DONATIONS = "donations"
    FAMILY_SUPPORT = "family_support"
    SALARY = "salary"



"""This module holds all POST simulation endpoints."""

# flake8: noqa: F401
from .simulate_account_topup import SimulateAccountTopup
from .simulate_transfer_state_update import SimulateTransferStateUpdate



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict
from pydantic_extra_types.currency_code import Currency

from pyrevolut.api.common import ModelBaseAmount, EnumPaymentDraftState
from pyrevolut.utils import Date


class RetrievePaymentDraft:
    """
    Get the information about a specific payment draft by ID.
    """

    ROUTE = "/1.0/payment-drafts/{payment_draft_id}"

    class Params(BaseModel):
        """
        The parameters for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        The response model for the endpoint.
        """

        class ModelPayment(BaseModel):
            """A payment draft."""

            class ModelAmount(ModelBaseAmount):
                """The amount of the payment draft."""

                pass

            class ModelReceiver(BaseModel):
                """The details of the transfer recipient.

                If the counterparty has multiple payment methods available
                (e.g. 2 accounts, or 1 account and 1 card), you must specify the
                account (account_id) or card (card_id) to which you want to transfer the money.
                """

                counterparty_id: Annotated[
                    UUID,
                    Field(description="The ID of the receiving counterparty."),
                ]
                account_id: Annotated[
                    UUID | None,
                    Field(
                        description="""
                        The ID of the receiving counterparty's account, which can be own account. 
                        Used for bank transfers.

                        If the counterparty has multiple payment methods available, use it to 
                        specify the account to which you want to send the money.
                        """
                    ),
                ] = None
                card_id: Annotated[
                    UUID | None,
                    Field(
                        description="""
                        The ID of the receiving counterparty's card. Used for card transfers.

                        If the counterparty has multiple payment methods available, use it to 
                        specify the card to which you want to send the money.
                        """
                    ),
                ] = None

            class ModelCurrentChargeOptions(BaseModel):
                """The explanation of conversion process"""

                model_config = ConfigDict(
                    populate_by_name=True,
                    from_attributes=True,
                )

                class ModelFrom(ModelBaseAmount):
                    """The source of the conversion"""

                    pass

                class ModelTo(ModelBaseAmount):
                    """The target of the conversion"""

                    pass

                class ModelFee(ModelBaseAmount):
                    """The fee of the conversion"""

                    pass

                from_: Annotated[
                    ModelFrom,
                    Field(alias="from", description="The source of the conversion."),
                ]
                to: Annotated[
                    ModelTo,
                    Field(description="The target of the conversion."),
                ]
                rate: Annotated[
                    float | None,
                    Field(description="The exchange rate used for the conversion."),
                ] = None
                fee: Annotated[
                    ModelFee | None,
                    Field(description="The fee of the conversion."),
                ] = None

            id: Annotated[UUID, Field(description="The ID of the payment draft.")]
            amount: Annotated[
                ModelAmount,
                Field(description="The amount of the payment draft."),
            ]
            currency: Annotated[
                Currency | None,
                Field(description="The currency of the payment draft."),
            ] = None
            account_id: Annotated[
                UUID,
                Field(description="The ID of the account to pay from."),
            ]
            receiver: Annotated[
                ModelReceiver,
                Field(
                    description="""
                    The details of the transfer recipient.

                    If the counterparty has multiple payment methods available 
                    (e.g. 2 accounts, or 1 account and 1 card), you must specify the 
                    account (account_id) or card (card_id) to which you want to transfer the money.
                    """
                ),
            ]
            state: Annotated[
                EnumPaymentDraftState,
                Field(description="The state of the payment draft."),
            ]
            reason: Annotated[
                str | None,
                Field(description="The reason for the payment draft state."),
            ] = None
            error_message: Annotated[
                str | None,
                Field(description="The description of the error message."),
            ] = None
            current_charge_options: Annotated[
                ModelCurrentChargeOptions,
                Field(description="The explanation of conversion process."),
            ]
            reference: Annotated[
                str | None,
                Field(description="The description of the transaction."),
            ] = None

        scheduled_for: Annotated[
            Date | None,
            Field(
                description="The scheduled date of the payment draft in ISO 8601 format."
            ),
        ] = None
        title: Annotated[
            str | None,
            Field(description="The title of the payment draft."),
        ] = None
        payments: Annotated[
            list[ModelPayment],
            Field(description="The list of payment drafts."),
        ]



"""The Simulations API is only available in the Sandbox environment.
It lets you simulate certain events that are otherwise only possible in the production environment,
such as your account's top-up and transfer state changes.

Reference link: https://developer.revolut.com/docs/business/simulations
"""

# flake8: noqa: F401
from .endpoint import EndpointSimulationsSync, EndpointSimulationsAsync



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from pyrevolut.utils import DateTime, Date
from pyrevolut.api.common import ModelBaseAmount, EnumCardState, EnumMerchantCategory


class ResourceCard(BaseModel):
    """
    Card resource model
    """

    class ModelSpendingLimits(BaseModel):
        """All spending limits set for the card."""

        class ModelSingle(ModelBaseAmount):
            """The limit for a single transaction."""

            pass

        class ModelDay(ModelBaseAmount):
            """The daily limit for transactions."""

            pass

        class ModelWeek(ModelBaseAmount):
            """The weekly limit for transactions."""

            pass

        class ModelMonth(ModelBaseAmount):
            """The monthly limit for transactions."""

            pass

        class ModelQuarter(ModelBaseAmount):
            """The quarterly limit for transactions."""

            pass

        class ModelYear(ModelBaseAmount):
            """The yearly limit for transactions."""

            pass

        class ModelAllTime(ModelBaseAmount):
            """The all-time limit for transactions."""

            pass

        single: Annotated[
            ModelSingle | None,
            Field(description="The limit for a single transaction."),
        ]
        day: Annotated[
            ModelDay | None,
            Field(description="The daily limit for transactions."),
        ]
        week: Annotated[
            ModelWeek | None,
            Field(description="The weekly limit for transactions."),
        ]
        month: Annotated[
            ModelMonth | None,
            Field(description="The monthly limit for transactions."),
        ]
        quarter: Annotated[
            ModelQuarter | None,
            Field(description="The quarterly limit for transactions."),
        ]
        year: Annotated[
            ModelYear | None,
            Field(description="The yearly limit for transactions."),
        ]
        all_time: Annotated[
            ModelAllTime | None,
            Field(description="The all-time limit for transactions."),
        ]

    id: Annotated[
        UUID,
        Field(description="The ID of the card."),
    ]
    last_digits: Annotated[
        str,
        Field(description="The last 4 digits of the card's PAN."),
    ]
    expiry: Annotated[
        Date,
        Field(description="The card expiration date."),
    ]
    state: Annotated[
        EnumCardState,
        Field(description="The state that the card is in."),
    ]
    label: Annotated[
        str | None,
        Field(description="The label of the card."),
    ]
    virtual: Annotated[
        bool,
        Field(
            description="Specifies whether the card is virtual (true) or physical (false)."
        ),
    ]
    accounts: Annotated[
        list[UUID],
        Field(description="The list of linked accounts."),
    ]
    categories: Annotated[
        list[EnumMerchantCategory] | None,
        Field(
            description="""
            The list of merchant categories that are available for card spending. 
            If not specified, all categories will be allowed.  
            """
        ),
    ]
    spending_limits: Annotated[
        ModelSpendingLimits | None,
        Field(description="All spending limits set for the card."),
    ]
    holder_id: Annotated[
        UUID | None,
        Field(
            description="""
            The ID of the team member who is the holder of the card. 
            If the card belongs to the business, this will be empty.
            """
        ),
    ]
    created_at: Annotated[
        DateTime,
        Field(description="The date and time the card was created in ISO 8601 format."),
    ]
    updated_at: Annotated[
        DateTime,
        Field(
            description="The date and time the card was last updated in ISO 8601 format."
        ),
    ]



from typing import Annotated

from pydantic import BaseModel, Field

from pyrevolut.utils import Date


class RetrieveSensitiveCardDetails:
    """
    Get sensitive details of a specific card, based on its ID.
    Requires the READ_SENSITIVE_CARD_DATA token scope.
    """

    ROUTE = "/1.0/cards/{card_id}/sensitive-details"

    class Params(BaseModel):
        """
        Query parameters for the endpoint.
        """

        pass

    class Response(BaseModel):
        """
        Response model for the endpoint.
        """

        pan: Annotated[
            str, Field(description="The PAN (Primary Account Number) of the card.")
        ]
        cvv: Annotated[
            str, Field(description="The CVV (Card Verification Value) of the card.")
        ]
        expiry: Annotated[Date, Field(description="The card expiration date.")]
